"""Pin definitions for the Odroid N2."""

from adafruit_blinka.microcontroller.amlogic.s922x import pin

GPIOX_0 = pin.GPIO476
GPIOX_1 = pin.GPIO477
GPIOX_2 = pin.GPIO478
GPIOX_3 = pin.GPIO479
GPIOX_4 = pin.GPIO480
GPIOX_5 = pin.GPIO481
GPIOX_6 = pin.GPIO482
GPIOX_7 = pin.GPIO483
GPIOX_8 = pin.GPIO484
GPIOX_9 = pin.GPIO485
GPIOX_10 = pin.GPIO486
GPIOX_11 = pin.GPIO487
GPIOX_12 = pin.GPIO488
GPIOX_13 = pin.GPIO489
GPIOX_14 = pin.GPIO490
GPIOX_15 = pin.GPIO491
GPIOX_16 = pin.GPIO492
GPIOX_17 = pin.GPIO493
GPIOX_18 = pin.GPIO494
GPIOX_19 = pin.GPIO495

GPIODV_24 = pin.GPIO493
GPIODV_25 = pin.GPIO494
GPIODV_26 = pin.GPIO474
GPIODV_27 = pin.GPIO475

GPIOA_4 = pin.GPIO464
GPIOA_12 = pin.GPIO472
GPIOA_13 = pin.GPIO473
GPIOA_14 = pin.GPIO474
GPIOA_15 = pin.GPIO475

GPIOA0_0 = pin.GPIO496
GPIOA0_1 = pin.GPIO497
GPIOA0_2 = pin.GPIO498
GPIOA0_3 = pin.GPIO499
GPIOA0_4 = pin.GPIO500
GPIOA0_5 = pin.GPIO501
GPIOA0_6 = pin.GPIO502
GPIOA0_7 = pin.GPIO503
GPIOA0_8 = pin.GPIO504
GPIOA0_9 = pin.GPIO505
GPIOA0_10 = pin.GPIO506
GPIOA0_11 = pin.GPIO507
GPIOA0_12 = pin.GPIO508
GPIOA0_13 = pin.GPIO509
GPIOA0_14 = pin.GPIO510
GPIOA0_15 = pin.GPIO511

for it in pin.i2cPorts:
    globals()["SCL" + str(it[0])] = it[1]
    globals()["SDA" + str(it[0])] = it[2]

# Set second i2c bus as default for backward compatibility.
SCL = pin.i2cPorts[1][1]
SDA = pin.i2cPorts[1][2]

SCLK = pin.SPI0_SCLK
MOSI = pin.SPI0_MOSI
MISO = pin.SPI0_MISO
SPI_CS0 = pin.GPIO486

D0 = GPIOX_3  # PIN_11
D1 = GPIOX_16  # PIN_12
D2 = GPIOX_4  # PIN_13
D3 = GPIOX_7  # PIN_15
D4 = GPIOX_0  # PIN_16
D5 = GPIOX_1  # PIN_18
D6 = GPIOX_2  # PIN_22
D7 = GPIOA_13  # PIN_7
D8 = GPIOX_17  # PIN_3
D9 = GPIOX_18  # PIN_5
D10 = GPIOX_10  # PIN_24
D11 = GPIOA_4  # PIN_26
D12 = GPIOX_8  # PIN_19
D13 = GPIOX_9  # PIN_21
D14 = GPIOX_11  # PIN_23
D15 = GPIOX_12  # PIN_8
D16 = GPIOX_13  # PIN_10
D21 = GPIOX_14  # PIN_29
D22 = GPIOX_15  # PIN_31
D23 = GPIOX_5  # PIN_33
D24 = GPIOX_6  # PIN_35
D26 = GPIOA_12  # PIN_32
D27 = GPIOX_19  # PIN_36
D30 = GPIOA_14  # PIN_27
D31 = GPIOA_15  # PIN_28



"""SPI Class for Binho Nova"""
from adafruit_blinka.microcontroller.nova import Connection


class SPI:
    """Custom SPI Class for Binho Nova"""

    MSB = 0
    BUFFER_PAYLOAD_MAX_LENGTH = 64
    WHR_PAYLOAD_MAX_LENGTH = 1024

    def __init__(self, clock):
        self._nova = Connection.getInstance()
        self._nova.setNumericalBase(10)
        self._nova.setOperationMode(0, "SPI")
        self._nova.setClockSPI(0, clock)
        self._nova.setModeSPI(0, 0)
        self._nova.setIOpinMode(0, "DOUT")
        self._nova.setIOpinMode(1, "DOUT")
        self._nova.beginSPI(0)
        self._novaCMDVer = "0"
        if hasattr(self._nova, "getCommandVer"):
            response = self._nova.getCommandVer().split(" ")
            if response[0] != "-NG":
                self._novaCMDVer = response[1]

        # Cpol and Cpha set by mode
        # Mode  Cpol Cpha
        #  0     0    0
        #  1     0    1
        #  2     1    0
        #  3     1    1

    def __del__(self):
        """Close Nova on delete"""
        self._nova.close()

    # pylint: disable=too-many-arguments,unused-argument
    def init(
        self,
        baudrate=1000000,
        polarity=0,
        phase=0,
        bits=8,
        firstbit=MSB,
        sck=None,
        mosi=None,
        miso=None,
    ):
        """Initialize the Port"""
        self._nova.setClockSPI(0, baudrate)
        self._nova.setModeSPI(0, (polarity << 1) | (phase))

    # pylint: enable=too-many-arguments,unused-argument

    @staticmethod
    def get_received_data(lineOutput):
        """Return any received data"""
        return lineOutput.split("RXD ")[1]

    @property
    def frequency(self):
        """Return the current frequency"""
        return self._nova.getClockSPI(0).split("CLK ")[1]

    def write(self, buf, start=0, end=None):
        """Write data from the buffer to SPI"""
        end = end if end else len(buf)
        payloadMaxLength = self.BUFFER_PAYLOAD_MAX_LENGTH
        if int(self._novaCMDVer) >= 1:
            payloadMaxLength = self.WHR_PAYLOAD_MAX_LENGTH
        chunks, rest = divmod(end - start, payloadMaxLength)

        for i in range(chunks):
            chunk_start = start + i * payloadMaxLength
            chunk_end = chunk_start + payloadMaxLength
            if int(self._novaCMDVer) >= 1:
                self._nova.writeToReadFromSPI(
                    0, True, False, chunk_end - chunk_start, buf[chunk_start:chunk_end]
                )
            else:
                self._nova.clearBuffer(0)
                self._nova.writeToBuffer(0, 0, buf[chunk_start:chunk_end])
                self._nova.transferBufferSPI(0, chunk_end - chunk_start + 1)
        if rest:
            if int(self._novaCMDVer) >= 1:
                self._nova.writeToReadFromSPI(0, True, False, rest, buf[-1 * rest :])
            else:
                self._nova.clearBuffer(0)
                self._nova.writeToBuffer(0, 0, buf[-1 * rest :])
                self._nova.transferBufferSPI(0, rest)

    def readinto(self, buf, start=0, end=None, write_value=0):
        """Read data from SPI and into the buffer"""
        end = end if end else len(buf)
        if int(self._novaCMDVer) >= 1:
            chunks, rest = divmod(end - start, self.WHR_PAYLOAD_MAX_LENGTH)
            i = 0
            for i in range(chunks):
                chunk_start = start + i * self.WHR_PAYLOAD_MAX_LENGTH
                chunk_end = chunk_start + self.WHR_PAYLOAD_MAX_LENGTH
                result = self._nova.writeToReadFromSPI(
                    0, False, True, chunk_end - chunk_start, write_value
                )
                if result != "-NG":
                    resp = result.split(" ")
                    resp = resp[2]
                    # loop over half of resp len as we're reading 2 chars at a time to form a byte
                    loops = int(len(resp) / 2)
                    for j in range(loops):
                        buf[(i * self.WHR_PAYLOAD_MAX_LENGTH) + start + j] = int(
                            resp[j * 2] + resp[j * 2 + 1], 16
                        )
                else:
                    raise RuntimeError(
                        "Received error response from Binho Nova, result = " + result
                    )
            if rest:
                result = self._nova.writeToReadFromSPI(
                    0, False, True, rest, write_value
                )
                if result != "-NG":
                    resp = result.split(" ")
                    resp = resp[2]

                    # loop over half of resp len as we're reading 2 chars at a time to form a byte
                    loops = int(len(resp) / 2)
                    for j in range(loops):
                        buf[(i * self.WHR_PAYLOAD_MAX_LENGTH) + start + j] = int(
                            resp[j * 2] + resp[j * 2 + 1], 16
                        )
                else:
                    raise RuntimeError(
                        "Received error response from Binho Nova, result = " + result
                    )
        else:
            for i in range(start, end):
                buf[start + i] = int(
                    self.get_received_data(self._nova.transferSPI(0, write_value))
                )

    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches
    def write_readinto(
        self, buffer_out, buffer_in, out_start=0, out_end=None, in_start=0, in_end=None
    ):
        """Perform a half-duplex write from buffer_out and then
        read data into buffer_in
        """
        out_end = out_end if out_end else len(buffer_out)
        in_end = in_end if in_end else len(buffer_in)
        readlen = in_end - in_start
        writelen = out_end - out_start
        if readlen > writelen:
            # resize out and pad with 0's
            tmp = bytearray(buffer_out)
            tmp.extend([0] * (readlen - len(buffer_out)))
            buffer_out = tmp

        if int(self._novaCMDVer) >= 1:
            chunks, rest = divmod(len(buffer_out), self.WHR_PAYLOAD_MAX_LENGTH)
            i = 0
            for i in range(chunks):
                chunk_start = out_start + i * self.WHR_PAYLOAD_MAX_LENGTH
                chunk_end = chunk_start + self.WHR_PAYLOAD_MAX_LENGTH
                result = self._nova.writeToReadFromSPI(
                    0,
                    True,
                    True,
                    chunk_end - chunk_start,
                    buffer_out[chunk_start:chunk_end],
                )

                if result != "-NG":
                    resp = result.split(" ")
                    resp = resp[2]

                    # loop over half of resp len as we're reading 2 chars at a time to form a byte
                    loops = int(len(resp) / 2)
                    for j in range(loops):
                        buffer_in[
                            (i * self.WHR_PAYLOAD_MAX_LENGTH) + in_start + j
                        ] = int(resp[j * 2] + resp[j * 2 + 1], 16)
                else:
                    raise RuntimeError(
                        "Received error response from Binho Nova, result = " + result
                    )
            if rest:
                result = self._nova.writeToReadFromSPI(
                    0, True, True, rest, buffer_out[-1 * rest :]
                )
                if result != "-NG":
                    resp = result.split(" ")
                    resp = resp[2]

                    # loop over half of resp len as we're reading 2 chars at a time to form a byte
                    loops = int(len(resp) / 2)
                    for j in range(loops):
                        buffer_in[
                            (i * self.WHR_PAYLOAD_MAX_LENGTH) + in_start + j
                        ] = int(resp[j * 2] + resp[j * 2 + 1], 16)
                else:
                    raise RuntimeError(
                        "Received error response from Binho Nova, result = " + result
                    )
            print(buffer_in)
        else:
            for data_out in buffer_out:
                data_in = int(
                    self.get_received_data(self._nova.transferSPI(0, data_out))
                )
                if i < readlen:
                    buffer_in[in_start + i] = data_in
                i += 1

    # pylint: enable=too-many-arguments,too-many-locals,too-many-branches



"""Atmel SAMA5 pin names"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

PD23 = Pin(119)
AD4 = PD23
PD21 = Pin(117)
AD2 = PD21
PD20 = Pin(116)
AD1 = PD20
PD24 = Pin(120)
AD5 = PD24
PD22 = Pin(118)
AD3 = PD22
PD19 = Pin(115)
AD0 = PD19
PA14 = Pin(14)
SPI0_SCLK = PA14
PA15 = Pin(15)
SPI0_MOSI = PA15
PA16 = Pin(16)
SPI0_MISO = PA16
PD2 = Pin(98)
UART1_RX = PD2
PD3 = Pin(99)
UART1_TX = PD3

PD13 = Pin(109)
PD31 = Pin(127)
PB0 = Pin(32)
PWM1 = PB0
PB7 = Pin(39)
PWM3 = PB7
PB1 = Pin(33)
PWML1 = PB1
PB5 = Pin(37)
PWM2 = PB5
PB3 = Pin(35)
PC0 = Pin(64)
TWI0_SCL = PC0
PB31 = Pin(63)
TWI0_SDA = PB31

i2cPorts = ((0, TWI0_SCL, TWI0_SDA),)
# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((0, SPI0_SCLK, SPI0_MOSI, SPI0_MISO),)
# ordered as uartId, txId, rxId
uartPorts = ((1, UART1_TX, UART1_RX),)
# SysFS pwm outputs, pwm channel and pin in first tuple
pwmOuts = (
    ((0, 1), PWM1),
    ((0, 2), PWM2),
    ((0, 3), PWM3),
)



"""Broadcom BCM283x pin names"""
import RPi.GPIO as GPIO

GPIO.setmode(GPIO.BCM)  # Use BCM pins D4 = GPIO #4
GPIO.setwarnings(False)  # shh!


class Pin:
    """Pins dont exist in CPython so...lets make our own!"""

    IN = 0
    OUT = 1
    LOW = 0
    HIGH = 1
    PULL_NONE = 0
    PULL_UP = 1
    PULL_DOWN = 2

    id = None
    _value = LOW
    _mode = IN

    def __init__(self, bcm_number):
        self.id = bcm_number

    def __repr__(self):
        return str(self.id)

    def __eq__(self, other):
        return self.id == other

    def init(self, mode=IN, pull=None):
        """Initialize the Pin"""
        if mode is not None:
            if mode == self.IN:
                self._mode = self.IN
                GPIO.setup(self.id, GPIO.IN)
            elif mode == self.OUT:
                self._mode = self.OUT
                GPIO.setup(self.id, GPIO.OUT)
            else:
                raise RuntimeError("Invalid mode for pin: %s" % self.id)
        if pull is not None:
            if self._mode != self.IN:
                raise RuntimeError("Cannot set pull resistor on output")
            if pull == self.PULL_UP:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            elif pull == self.PULL_DOWN:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            else:
                raise RuntimeError("Invalid pull for pin: %s" % self.id)

    def value(self, val=None):
        """Set or return the Pin Value"""
        if val is not None:
            if val == self.LOW:
                self._value = val
                GPIO.output(self.id, val)
            elif val == self.HIGH:
                self._value = val
                GPIO.output(self.id, val)
            else:
                raise RuntimeError("Invalid value for pin")
            return None
        return GPIO.input(self.id)


# Pi 1B rev1 only?
D0 = Pin(0)
D1 = Pin(1)

D2 = Pin(2)
SDA = Pin(2)
D3 = Pin(3)
SCL = Pin(3)

D4 = Pin(4)
D5 = Pin(5)
D6 = Pin(6)

D7 = Pin(7)
CE1 = Pin(7)
D8 = Pin(8)
CE0 = Pin(8)
D9 = Pin(9)
MISO = Pin(9)
D10 = Pin(10)
MOSI = Pin(10)
D11 = Pin(11)
SCLK = Pin(11)  # Raspberry Pi naming
SCK = Pin(11)  # CircuitPython naming

D12 = Pin(12)
D13 = Pin(13)

D14 = Pin(14)
TXD = Pin(14)
D15 = Pin(15)
RXD = Pin(15)

D16 = Pin(16)
D17 = Pin(17)
D18 = Pin(18)
D19 = Pin(19)
MISO_1 = Pin(19)
D20 = Pin(20)
MOSI_1 = Pin(20)
D21 = Pin(21)
SCLK_1 = Pin(21)
SCK_1 = Pin(21)
D22 = Pin(22)
D23 = Pin(23)
D24 = Pin(24)
D25 = Pin(25)
D26 = Pin(26)
D27 = Pin(27)
D28 = Pin(28)
D29 = Pin(29)
D30 = Pin(30)
D31 = Pin(31)
D32 = Pin(32)
D33 = Pin(33)
D34 = Pin(34)
D35 = Pin(35)
D36 = Pin(36)
D37 = Pin(37)
D38 = Pin(38)
D39 = Pin(39)
D40 = Pin(40)
MISO_2 = Pin(40)
D41 = Pin(41)
MOSI_2 = Pin(41)
D42 = Pin(42)
SCLK_2 = Pin(42)
SCK_2 = Pin(43)
D43 = Pin(43)
D44 = Pin(44)
D45 = Pin(45)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = (
    (0, SCLK, MOSI, MISO),
    (1, SCLK_1, MOSI_1, MISO_1),
    (2, SCLK_2, MOSI_2, MISO_2),
)

# ordered as uartId, txId, rxId
uartPorts = ((1, TXD, RXD),)

# These are the known hardware I2C ports / pins.
# For software I2C ports created with the i2c-gpio overlay, see:
#     https://github.com/adafruit/Adafruit_Python_Extended_Bus
i2cPorts = (
    (1, SCL, SDA),
    (0, D1, D0),  # both pi 1 and pi 2 i2c ports!
)



"""Pin definitions for the NanoPi Duo2."""
# Enable I2C0, UART1, and SPI by adding the following lines to /boot/armbianEnv.txt
#    overlays=usbhost2 usbhost3 spi-spidev uart1 i2c0
#    param_spidev_spi_bus=0

from adafruit_blinka.microcontroller.allwinner.h3 import pin

# Left GPIO
PG11 = pin.PG11

# I2C
SDA = pin.PA12
SCL = pin.PA11

# SPI
SCLK = pin.PA14
MOSI = pin.PA15
MISO = pin.PA16
CE0 = pin.PA13

# Serial UART
UART1_TX = pin.PG6
UART1_RX = pin.PG7



"""Allwinner H5 pin names"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

PA0 = Pin((1, 0))
UART2_TX = PA0
PA1 = Pin((1, 1))
UART2_RX = PA1
PA2 = Pin((1, 2))
PA3 = Pin((1, 3))
PA6 = Pin((1, 6))
PA7 = Pin((1, 7))
PA8 = Pin((1, 8))
PA9 = Pin((1, 9))
PA10 = Pin((1, 10))
PA11 = Pin((1, 11))
TWI0_SCL = PA11
PA12 = Pin((1, 12))
TWI0_SDA = PA12
PA13 = Pin((1, 13))
UART3_TX = PA13
PA14 = Pin((1, 14))
UART3_RX = PA14
SPI1_SCLK = PA14
PA15 = Pin((1, 15))
SPI1_MOSI = PA15
PA16 = Pin((1, 16))
SPI1_MISO = PA16
PA17 = Pin((1, 17))
PA18 = Pin((1, 18))
PA19 = Pin((1, 19))
PA20 = Pin((1, 20))
PA21 = Pin((1, 21))

PC0 = Pin((1, 64))
SPI0_MOSI = PC0
PC1 = Pin((1, 65))
SPI0_MISO = PC1
PC2 = Pin((1, 66))
SPI0_SCLK = PC2
PC3 = Pin((1, 67))
SPI0_CS = PC3
PC4 = Pin((1, 68))
PC7 = Pin((1, 71))

PD14 = Pin((1, 110))

PG6 = Pin((1, 198))
UART1_TX = PG6
PG7 = Pin((1, 199))
UART1_RX = PG7
PG8 = Pin((1, 200))
PG9 = Pin((1, 201))
PG10 = Pin((1, 202))
PG11 = Pin((1, 203))
PG12 = Pin((1, 204))
PG13 = Pin((1, 205))

PL0 = Pin((0, 0))
PL1 = Pin((0, 1))

i2cPorts = ((0, TWI0_SCL, TWI0_SDA),)
# ordered as spiId, sckId, mosiId, misoId
spiPorts = (
    (0, SPI0_SCLK, SPI0_MOSI, SPI0_MISO),
    (1, SPI1_SCLK, SPI1_MOSI, SPI1_MISO),
)
# ordered as uartId, txId, rxId
uartPorts = ((3, UART3_TX, UART3_RX),)



"""I2C Class for NXP LPC4330"""
from greatfet import GreatFET


class I2C:
    """Custom I2C Class for NXP LPC4330"""

    # pylint: disable=unused-argument
    def __init__(self, *, frequency=100000):
        self._gf = GreatFET()

    def scan(self):
        """Perform an I2C Device Scan"""
        return [index for index, dev in enumerate(self._gf.i2c.scan()) if dev[0]]

    def writeto(self, address, buffer, *, start=0, end=None, stop=True):
        """Write data from the buffer to an address"""
        if end is None:
            end = len(buffer)
        self._gf.i2c.write(address, buffer[start:end])

    def readfrom_into(self, address, buffer, *, start=0, end=None, stop=True):
        """Read data from an address and into the buffer"""
        if end is None:
            end = len(buffer)
        readin = self._gf.i2c.read(address, end - start)
        for i in range(end - start):
            buffer[i + start] = readin[i]

    # pylint: enable=unused-argument

    def writeto_then_readfrom(
        self,
        address,
        buffer_out,
        buffer_in,
        *,
        out_start=0,
        out_end=None,
        in_start=0,
        in_end=None,
        stop=False
    ):
        """Write data from buffer_out to an address and then
        read data from an address and into buffer_in
        """
        self.writeto(address, buffer_out, start=out_start, end=out_end, stop=stop)
        self.readfrom_into(address, buffer_in, start=in_start, end=in_end, stop=stop)



"""Pin definitions for the Raspberry Pi Model B Revision 1."""

from adafruit_blinka.microcontroller.bcm283x import pin

D0 = pin.D0
D1 = pin.D1
SDA = pin.D0
SCL = pin.D1
D4 = pin.D4

D7 = pin.D7
CE1 = pin.D7
D8 = pin.D8
CE0 = pin.D8
D9 = pin.D9
MISO = pin.D9
D10 = pin.D10
MOSI = pin.D10
D11 = pin.D11
SCLK = pin.D11
SCK = pin.D11

D14 = pin.D14
TXD = pin.D14
D15 = pin.D15
RXD = pin.D15
# create alias for most of the examples
TX = pin.D14
RX = pin.D15

D17 = pin.D17
D18 = pin.D18
D21 = pin.D21
D22 = pin.D22
D23 = pin.D23
D24 = pin.D24
D25 = pin.D25



"""Pin definitions for the Coral Dev Board Mini."""

from adafruit_blinka.microcontroller.mt8167 import pin

# Board name = RPI name [= alias] = pin name
GPIO22 = D4 = pin.GPIO22  # Pin 7
GPIO9 = D17 = pin.GPIO9  # Pin 11
GPIO36 = D18 = pin.GPIO36  # Pin 12
GPIO10 = D27 = pin.GPIO10  # Pin 13
GPIO0 = D23 = pin.GPIO0  # Pin 16
GPIO1 = D24 = pin.GPIO1  # Pin 18
GPIO7 = D25 = pin.GPIO7  # Pin 22
GPIO8 = D7 = pin.GPIO8  # Pin 26
GPIO37 = D19 = pin.GPIO37  # Pin 35
GPIO13 = D16 = pin.GPIO13  # Pin 36
GPIO45 = D26 = pin.GPIO45  # Pin 37
GPIO38 = D20 = pin.GPIO38  # Pin 38
GPIO39 = D21 = pin.GPIO39  # Pin 40

I2C1_SDA = D2 = SDA1 = pin.I2C1_SDA  # Pin 3
I2C1_SCL = D3 = SCL1 = pin.I2C1_SCL  # Pin 5
I2C2_SDA = D0 = SDA2 = pin.I2C2_SDA  # Pin 27
I2C2_SCL = D1 = SCL2 = pin.I2C2_SCL  # Pin 28

PWM_A = D12 = pin.PWM_A  # Pin 32
PWM_B = D13 = pin.PWM_B  # Pin 33
PWM_C = D22 = pin.PWM_C  # Pin 15

SPI_MO = D10 = MOSI = pin.SPI_MO  # Pin 19
SPI_MI = D9 = MISO = pin.SPI_MI  # Pin 21
SPI_CLK = D11 = SCLK = pin.SPI_CLK  # Pin 23
SPI_CSB = D8 = CS0 = pin.SPI_CSB  # Pin 24

# UART currently not supported
GPIO63 = D14 = pin.GPIO63  # UART0_TX, Pin 8
GPIO62 = D15 = pin.GPIO62  # UART0_RX, Pin 10
GPIO65 = D5 = pin.GPIO65  # UART1_TX, Pin 29
GPIO64 = D6 = pin.GPIO64  # UART1_RX, Pin 31



"""Pin definitions for the FTDI FT232H"""
from adafruit_blinka.microcontroller.ft232h import pin

D4 = pin.D4
D5 = pin.D5
D6 = pin.D6
D7 = pin.D7
C0 = pin.C0
C1 = pin.C1
C2 = pin.C2
C3 = pin.C3
C4 = pin.C4
C5 = pin.C5
C6 = pin.C6
C7 = pin.C7

SDA = pin.SDA
SCL = pin.SCL

SCK = pin.SCK
SCLK = pin.SCLK
MOSI = pin.MOSI
MISO = pin.MISO



"""Generic Linux SPI class using PureIO's SPI class"""
import Adafruit_PureIO.spi as spi
from adafruit_blinka.agnostic import detector


class SPI:
    """SPI Class"""

    MSB = 0
    LSB = 1
    CPHA = 1
    CPOL = 2

    baudrate = 100000
    mode = 0
    bits = 8

    def __init__(self, portid):
        if isinstance(portid, tuple):
            self._spi = spi.SPI(device=portid)
        else:
            self._spi = spi.SPI(device=(portid, 0))
        self.clock_pin = None
        self.mosi_pin = None
        self.miso_pin = None
        self.chip = None

    # pylint: disable=too-many-arguments,unused-argument
    def init(
        self,
        baudrate=100000,
        polarity=0,
        phase=0,
        bits=8,
        firstbit=MSB,
        sck=None,
        mosi=None,
        miso=None,
    ):
        """Initialize SPI"""
        mode = 0
        if polarity:
            mode |= self.CPOL
        if phase:
            mode |= self.CPHA
        self.baudrate = baudrate
        self.mode = mode
        self.bits = bits
        self.chip = detector.chip

        # Pins are not used
        self.clock_pin = sck
        self.mosi_pin = mosi
        self.miso_pin = miso

    # pylint: enable=too-many-arguments,unused-argument

    # pylint: disable=unnecessary-pass
    def set_no_cs(self):
        """Setting so that SPI doesn't automatically set the CS pin"""
        # No kernel seems to support this, so we're just going to pass
        pass

    # pylint: enable=unnecessary-pass

    @property
    def frequency(self):
        """Return the current baudrate"""
        return self.baudrate

    def write(self, buf, start=0, end=None):
        """Write data from the buffer to SPI"""
        if not buf:
            return
        if end is None:
            end = len(buf)
        try:
            # self._spi.open(self._port, 0)
            self.set_no_cs()
            self._spi.max_speed_hz = self.baudrate
            self._spi.mode = self.mode
            self._spi.bits_per_word = self.bits
            self._spi.writebytes(buf[start:end])
            # self._spi.close()
        except FileNotFoundError:
            print("Could not open SPI device - check if SPI is enabled in kernel!")
            raise

    def readinto(self, buf, start=0, end=None, write_value=0):
        """Read data from SPI and into the buffer"""
        if not buf:
            return
        if end is None:
            end = len(buf)
        try:
            # self._spi.open(self._port, 0)
            # self.set_no_cs()
            self._spi.max_speed_hz = self.baudrate
            self._spi.mode = self.mode
            self._spi.bits_per_word = self.bits
            data = self._spi.transfer([write_value] * (end - start))
            for i in range(end - start):  # 'readinto' the given buffer
                buf[start + i] = data[i]
            # self._spi.close()
        except FileNotFoundError:
            print("Could not open SPI device - check if SPI is enabled in kernel!")
            raise

    # pylint: disable=too-many-arguments
    def write_readinto(
        self, buffer_out, buffer_in, out_start=0, out_end=None, in_start=0, in_end=None
    ):
        """Perform a half-duplex write from buffer_out and then
        read data into buffer_in
        """
        if not buffer_out or not buffer_in:
            return
        if out_end is None:
            out_end = len(buffer_out)
        if in_end is None:
            in_end = len(buffer_in)
        if out_end - out_start != in_end - in_start:
            raise RuntimeError("Buffer slices must be of equal length.")
        try:
            # self._spi.open(self._port, 0)
            # self.set_no_cs()
            self._spi.max_speed_hz = self.baudrate
            self._spi.mode = self.mode
            self._spi.bits_per_word = self.bits
            data = self._spi.transfer(list(buffer_out[out_start : out_end + 1]))
            for i in range((in_end - in_start)):
                buffer_in[i + in_start] = data[i]
            # self._spi.close()
        except FileNotFoundError:
            print("Could not open SPI device - check if SPI is enabled in kernel!")
            raise

    # pylint: enable=too-many-arguments



"""
Pin definitions for the Beaglebone PocketBeagle.

based on
https://github.com/beagleboard/pocketbeagle/wiki/System-Reference-Manual#figure-42-expansion-header-popular-functions---color-coded
"""
from adafruit_blinka.microcontroller.am335x import pin

# initial pins, to mimic bonescript demo
# PocketBeagle
# P1_1 = SYS VIN        # VIN_AC
P1_2 = pin.P1_2  # GPIO2_23 - GPIO_87
P1_3 = pin.P1_3  # USB1_VBUS_OUT - (silkscreen: USB1 V_EN)
P1_4 = pin.P1_4  # GPIO2_25 - GPIO_89
# P1_5 = USB VBUS       # USB1_VBUS_IN
P1_6 = pin.P1_6  # SPI0_CS0 - GPIO_5
# P1_7 = USB VIN        # VIN-USB
P1_8 = pin.P1_8  # SPI0_SCLK - GPIO_2
# P1_9 = USB1 DN        # USB1-DN
P1_10 = pin.P1_10  # SPI0_D0 - GPIO_3
# P1_11 = USB1 DP       # USB1-DP
P1_12 = pin.P1_12  # SPI0_D1 - GPIO_4
# P1_13 = USB1 ID       # USB1-ID
# P1_14 = SYS 3.3V      # VOUT-3.3V
# P1_15 = SYS GND       # GND
# P1_16 = SYS GND       # GND
# P1_17 = AIN 1.8V REF- # VREFN
# P1_18 = AIN 1.8V REF+ # VREFP
P1_19 = pin.P1_19  # AIN0
P1_20 = pin.P1_20  # GPIO0_20 - GPIO_20
P1_21 = pin.P1_21  # AIN1
# P1_22 = SYS GND       # GND
P1_23 = pin.P1_23  # AIN2
# P1_22 = SYS VOUT      # VOUT-5V
P1_25 = pin.P1_25  # AIN3
P1_26 = pin.P1_26  # I2C2_SDA - GPIO_12
P1_27 = pin.P1_27  # AIN4
P1_28 = pin.P1_28  # I2C2_SCL - GPIO_13
P1_29 = pin.P1_29  # GPIO3_21 - GPIO_117
P1_30 = pin.P1_30  # UART0_TXD - GPIO_43
P1_31 = pin.P1_31  # GPIO3_18 - GPIO_114
P1_32 = pin.P1_32  # UART0_RXD - GPIO_42
P1_33 = pin.P1_33  # GPIO3_15 - GPIO_111 - EHRPWM0B (ehrpwm.0:1)
P1_34 = pin.P1_34  # GPIO0_26 - GPIO_26
P1_35 = pin.P1_35  # GPIO2_24 - GPIO_88
P1_36 = pin.P1_36  # EHRPWM0A - GPIO_110 - EHRPWM0A (ehrpwm.0:0)


P2_1 = pin.P2_1  # EHRPWM1A - GPIO_50
P2_2 = pin.P2_2  # GPIO1_27 - GPIO_59
P2_3 = pin.P2_3  # GPIO0_23 - GPIO_23
P2_4 = pin.P2_4  # GPIO1_26 - GPIO_58
P2_5 = pin.P2_5  # UART4_RXD - GPIO_30
P2_6 = pin.P2_6  # GPIO1_25 - GPIO_57
P2_7 = pin.P2_7  # UART4_TXD - GPIO_31
P2_8 = pin.P2_8  # GPIO1_28 - GPIO_60
P2_9 = pin.P2_9  # I2C1_SCL - GPIO_15
P2_10 = pin.P2_10  # GPIO1_20 - GPIO_52
P2_11 = pin.P2_11  # I2C1_SDA - GPIO_14
# P2_12 = SYS  PWR BTN  # POWER_BUTTON
# P2_13 = SYS VOUT      # VOUT-5V
# P2_14 = BAT VIN       # BAT-VIN
# P2_15 = SYS GND       # GND
# P2_16 = BAT TEMP      # BAT-TEMP
P2_17 = pin.P2_17  # GPIO2_1 - GPIO_65
P2_18 = pin.P2_18  # GPIO1_15 - GPIO_47
P2_19 = pin.P2_19  # GPIO0_27 - GPIO_27
P2_20 = pin.P2_20  # GPIO2_0 - GPIO_64
# P2_21 = SYS GND       # GND
P2_22 = pin.P2_22  # GPIO1_14 - GPIO_46
# P2_23 = SYS 3.3V      # VOUT-3.3V
P2_24 = pin.P2_24  # GPIO1_12 - GPIO_44
P2_25 = pin.P2_25  # SPI1_CS0 - GPIO_41
# P2_26 = SYS NRST      # RESET#
P2_27 = pin.P2_27  # SPI1_D0 - GPIO_40
P2_28 = pin.P2_28  # GPIO3_20 - GPIO_116
P2_29 = pin.P2_29  # SPI1_SCLK - GPIO_7
P2_30 = pin.P2_30  # GPIO3_17 - GPIO_113
P2_31 = pin.P2_31  # SPI1_CS1 - GPIO_19
P2_32 = pin.P2_32  # GPIO3_16 - GPIO_112
P2_33 = pin.P2_33  # GPIO1_13 - GPIO_45
P2_34 = pin.P2_34  # GPIO3_19 - GPIO_115
P2_35 = pin.P2_35  # GPIO2_22 - GPIO_86
P2_36 = pin.P2_36  # AIN7

# common to all beagles
LED_USR0 = pin.USR0
LED_USR1 = pin.USR1
LED_USR2 = pin.USR2
LED_USR3 = pin.USR3

##########
# Refer to header default pin modes
# https://raw.githubusercontent.com/wiki/beagleboard/pocketbeagle/images/PocketBeagle_pinout.png

# I2C1 pins
SDA_1 = pin.I2C1_SDA  # P2_11 data signal
SCL_1 = pin.I2C1_SCL  # P2_9 clock signal
# for example compatibility we create a alias
SDA = SDA_1
SCL = SCL_1

# I2C2 pins
SDA_2 = pin.I2C2_SDA  # P1_26 data signal
SCL_2 = pin.I2C2_SCL  # P1_28 clock signal

# SPI0 pins
CE0 = pin.SPI0_CS0  # P1_6 - enables peripheral device
SCLK = pin.SPI0_SCLK  # P1_12 - outputs data to peripheral device
MOSI = pin.SPI0_D1  # P1_10 - receives data from peripheral device
MISO = pin.SPI0_D0  # P1_8 - outputs clock signal
# CircuitPython naming convention for SPI Clock
SCK = SCLK

# SPI1 pins
CE0_1 = pin.SPI1_CS1  # P2_31 - enables peripheral device
SCLK_1 = pin.SPI1_SCLK  # P2_25 - outputs data to peripheral device
MOSI_1 = pin.SPI1_D1  # P2_27 - receives data from peripheral device
MISO_1 = pin.SPI1_D0  # P2_29 - outputs clock signal
# CircuitPython naming convention for SPI Clock
SCK_1 = SCLK_1


# UART0
TX_0 = pin.UART0_TXD  # P1_30
RX_0 = pin.UART0_RXD  # P1_32
# create alias for most of the examples
TX = TX_0
RX = RX_0

# UART2
# pins already in use by SPI0
# TX_2 = pin.UART2_TXD    # P1_8
# RX_2 = pin.UART2_RXD    # P1_10

# UART4
TX_4 = pin.UART4_TXD  # P2_7
RX_4 = pin.UART4_RXD  # P2_5



"""Pentium N3710 (Braswell core SOC) pin names
   i2c and GPIO can be accessed through Blinka.
   For i2c use IC20_SCL, IC20-SDA and IC21-SCL, IC21-SDA in the i2c(<sdl, sda>) calls.
   For UART use pyserial"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

# gpiochip3
GPIO_243 = Pin((3, 15))
GPIO_246 = Pin((3, 18))
GPIO_247 = Pin((3, 19))
GPIO_249 = Pin((3, 21))
GPIO_250 = Pin((3, 22))
GPIO_253 = Pin((3, 25))
GPIO_273 = Pin((3, 45))
GPIO_275 = Pin((3, 47))
GPIO_276 = Pin((3, 48))
GPIO_278 = Pin((3, 50))
GPIO_279 = Pin((3, 51))
GPIO_280 = Pin((3, 52))
GPIO_307 = Pin((3, 79))

SDIO_D3 = SDMMC2_D3 = GPIO_243
SDIO_DI = SDMMC2_D1 = GPIO_246
SDIO_CLK = SDMMC2_CLK = GPIO_247
SDIO_D2 = SDMMC2_D2 = GPIO_249
SDIO_CMD = SDMMC2_CMD = GPIO_250
SDIO_D0 = SDMMC2_D0 = GPIO_253

MF_LPC_AD2 = GPIO_273
MF_LPC_AD0 = GPIO_275
LPC_FRAMEB = GPIO_276
MF_LPC_AD3 = GPIO_278
MF_LPC_CLKOUT0 = GPIO_279
MF_LPC_AD1 = GPIO_280
ILB_SERIRQ = GPIO_307

# ggpiochip1
GPIO_358 = Pin((1, 17))
GPIO_SUS3 = SDIO_WAKE = GPIO_358


# gpiochip0
GPIO_490 = Pin((0, 76))
GPIO_492 = Pin((0, 78))

SATA_GP1 = TS_INT = GPIO_490
SATA_GP2 = TS_RST = GPIO_492


# not general gpio on chip 0
# use pyserial not blinka.  These are only included for completeness

UART1_RXD = Pin((0, 16))
UART1_TXD = Pin((0, 20))
UART1_RTS = Pin((0, 15))
UART1_CTS = Pin((0, 18))

UART2_RXD = Pin((0, 17))
UART2_TXD = Pin((0, 21))
UART2_RTS = Pin((0, 19))
UART2_CTS = Pin((0, 22))

GPIO_429 = UART1_RTS
GPIO_430 = UART1_RXD
GPIO_431 = UART2_RXD
GPIO_432 = UART1_CTS
GPIO_434 = UART1_TXD
GPIO_435 = UART2_TXD
GPIO_436 = UART2_CTS

# i2c use these addresses when accessing i2c from Blinka.  You can also access
# the i2c useing smbus
I2C0_SDA = Pin((0, 61))  # IC21 on diagram, port 0 in hardware manual
I2C0_SCL = Pin((0, 65))

I2C1_SDA = TS_I2C_SDA = Pin((0, 45))  # I2C2 on diagram, port 5 in hardware manual
I2C1_SCL = TS_I2C_SCL = Pin((0, 48))


GPIO_469 = I2C1_SDA  # I2C2 on diagram
GPIO_472 = I2C1_SCL
GPIO_475 = I2C0_SDA  # I2C1 on diagram
GPIO_479 = I2C0_SCL

# ordered as i2cId, sclId, sdaId
i2cPorts = ((0, I2C0_SCL, I2C0_SDA), (1, I2C1_SCL, I2C1_SDA))



"""Allwinner A33 pin names"""

from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

PB0 = Pin(32)  # PB0/UART2_TX/UART0_TX/PB_EINT0
UART2_TX = PB0
PB1 = Pin(33)  # PB1/UART2_RX/UART0_RX/PB_EINT1
UART2_RX = PB1

PC0 = Pin(64)  # PC0/ND_WE/SPI0_MOSI
PC1 = Pin(65)  # PC1/ND_ALE/SPI0_MISO
PC2 = Pin(66)  # PC2/ND_CLE/SPI0_CLK

PH4 = Pin(228)  # PH4/TWI1_SCK
TWI1_SCL = PH4
PH5 = Pin(229)  # PH5/TWI1_SDA
TWI1_SDA = PH5


PH6 = Pin(230)  # PH6/SPI0_CS/UART3_TX
UART3_TX = PH6
SPI0_CS = PH6

PH7 = Pin(231)  # PH7/SPI0_CLK/UART3_RX
UART3_RX = PH7
SPI0_SCLK = PH7

PH8 = Pin(232)  # PH8/SPI0_MOSI/UART3_RTS
UART3_RTS = PH8
SPI0_MOSI = PH8

PH9 = Pin(233)  # PH9/SPI0_MISO/UART3_CTS
UART3_CTS = PH9
SPI0_MISO = PH9


# ordered as i2cId, sclId, sdaId
i2cPorts = ((0, TWI1_SCL, TWI1_SDA),)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((0, SPI0_SCLK, SPI0_MOSI, SPI0_MISO),)

# ordered as uartId, txId, rxId
uartPorts = (
    (2, UART2_TX, UART2_RX),
    (3, UART3_TX, UART3_RX),
)



"""Pin definitions for the Orange Pi Zero 2."""

from adafruit_blinka.microcontroller.allwinner.h616 import pin

PH5 = pin.PH5
SDA = pin.PH5
PH4 = pin.PH4
SCL = pin.PH4
PC9 = pin.PC9

PH3 = pin.PH3
UART5_RX = pin.PH3
PH2 = pin.PH2
UART5_TX = pin.PH2
PC8 = pin.PC8

PC10 = pin.PC10
PC11 = pin.PC11
PC6 = pin.PC6
PC5 = pin.PC5

PH9 = pin.PH9
SPI1_CS = pin.PH9
PH6 = pin.PH6
SPI1_CLK = pin.PH6
PC7 = pin.PC7

PC14 = pin.PC14

PC15 = pin.PC15


SCLK = pin.PH6
MOSI = pin.PH7
MISO = pin.PH8



"""Pin definitions for the Giant Board"""
from adafruit_blinka.microcontroller.sama5 import pin

PD23 = pin.PD23
AD4 = pin.PD23
PD21 = pin.PD21
AD2 = pin.PD21
PD20 = pin.PD20
AD1 = pin.PD20
PD24 = pin.PD24
AD5 = pin.PD24
PD22 = pin.PD22
AD3 = pin.PD22
PD19 = pin.PD19
AD0 = pin.PD19
PA14 = pin.PA14
SCK = pin.PA14
SCLK = pin.PA14
PA15 = pin.PA15
MOSI = pin.PA15
PA16 = pin.PA16
MISO = pin.PA16
PD2 = pin.PD2
RX = pin.PD2
PD3 = pin.PD3
TX = pin.PD3

PD13 = pin.PD13
PD31 = pin.PD31
PB0 = pin.PB0
PWM1 = pin.PB0
PB7 = pin.PB7
PWM3 = pin.PB7
PB1 = pin.PB1
PWML1 = pin.PB1
PB5 = pin.PB5
PWM2 = pin.PB5
PB3 = pin.PB3
PC0 = pin.PC0
SCL = pin.PC0
PB31 = pin.PB31
SDA = pin.PB31



"""NXP IMX8M pin names"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

I2C2_SCL = Pin(144)  # GPIO5_IO16
I2C2_SDA = Pin(145)  # GPIO5_IO17

I2C3_SCL = Pin(146)  # GPIO5_IO18
I2C3_SDA = Pin(147)  # GPIO5_IO19


PWM1 = Pin((0, 1))  # GPIO1_IO01
PWM2 = Pin((0, 13))  # GPIO1_IO13
PWM3 = Pin((0, 14))  # GPIO1_IO14

GPIO6 = Pin((0, 6))  # GPIO1_IO6
GPIO7 = Pin((0, 7))  # GPIO1_IO7
GPIO8 = Pin((0, 8))  # GPIO1_IO8
GPIO73 = Pin((2, 9))  # GPIO3_IO9
GPIO77 = Pin((2, 13))  # GPIO3_IO13
GPIO138 = Pin((4, 10))  # GPIO5_IO10
GPIO141 = Pin((4, 13))  # GPIO5_IO13

ECSPI1_MISO = Pin(136)  # GPIO5_IO8
ECSPI1_MOSI = Pin(135)  # GPIO5_IO7
ECSPI1_SCLK = Pin(134)  # GPIO5_IO6
ECSPI1_SS0 = Pin(133)  # GPIO5_IO9


i2cPorts = (
    (1, I2C2_SCL, I2C2_SDA),
    (2, I2C3_SCL, I2C3_SDA),
)
# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((0, ECSPI1_SCLK, ECSPI1_MOSI, ECSPI1_MISO),)
# SysFS pwm outputs, pwm channel and pin in first tuple
pwmOuts = (
    ((0, 0), PWM1),
    ((1, 0), PWM2),
    ((2, 0), PWM3),
)

# UART1_TXD/RXD on /dev/ttymxc0
# UART3_TXD/RXD not available (?)



"""Hifive Unleashed pin names"""

from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

UART0_TXD = Pin(5)
UART0_RXD = Pin(7)
SPI0_SCLK = Pin(8)
SPI0_DIN = Pin(10)
UART1_TXD = Pin(11)
SPI0_CS = Pin(12)
UART1_RXD = Pin(13)
SPI0_DOUT = Pin(14)
I2C0_SCL = Pin(15)
I2C0_SDA = Pin(17)
GPIO0 = Pin(23)
GPIO1 = Pin(24)
GPIO2 = Pin(25)
GPIO3 = Pin(26)
GPIO4 = Pin(27)
GPIO5 = Pin(28)
GPIO6 = Pin(29)
GPIO7 = Pin(30)
GPIO8 = Pin(31)
GPIO9 = Pin(32)
GPIO15 = Pin(33)

# ordered as spiId, sckId, mosiId, misoId
SPI_PORTS = (1, SPI0_SCLK, SPI0_DOUT, SPI0_DIN)

# ordered as uartId, txId, rxId
UART_PORTS = (
    (0, UART0_TXD, UART0_RXD),
    # (0, GPIO15, GPIO13)
    (1, UART1_TXD, UART1_RXD),
)

# ordered as i2cId, sclId, sdaId
I2C_PORTS = (0, I2C0_SDA, I2C0_SCL)



"""SPI Class for NXP LPC4330"""
from greatfet import GreatFET


class SPI:
    """Custom I2C Class for NXP LPC4330"""

    MSB = 0

    def __init__(self):
        self._gf = GreatFET()
        self._frequency = None
        self.buffer_size = 255
        self._mode = 0
        self._spi = None
        self._presets = {
            204000: (100, 9),
            408000: (100, 4),
            680000: (100, 2),
            1020000: (100, 1),
            2040000: (50, 1),
            4250000: (24, 1),
            8500000: (12, 1),
            12750000: (8, 1),
            17000000: (6, 1),
            20400000: (2, 4),
            25500000: (4, 1),
            34000000: (2, 2),
            51000000: (2, 1),
            102000000: (2, 0),
        }

    # pylint: disable=too-many-arguments,unused-argument
    def init(
        self,
        baudrate=100000,
        polarity=0,
        phase=0,
        bits=8,
        firstbit=MSB,
        sck=None,
        mosi=None,
        miso=None,
    ):
        """Initialize the Port"""
        # Figure out the mode based on phase and polarity
        polarity = int(polarity)
        phase = int(phase)
        self._mode = (polarity << 1) | phase

        # Using API due to possible interface change
        self._spi = self._gf.apis.spi
        # Check baudrate against presets and adjust to the closest one
        if self._frequency is None:
            preset = self._find_closest_preset(baudrate)
        else:
            preset = self._presets[self._frequency]
        clock_prescale_rate, serial_clock_rate = preset
        self._spi.init(serial_clock_rate, clock_prescale_rate)

        # Set the polarity and phase (the "SPI mode").
        self._spi.set_clock_polarity_and_phase(self._mode)

    # pylint: enable=too-many-arguments

    def _find_closest_preset(self, target_frequency):
        """Loop through self._frequencies and find the closest
        setting. Return the preset values and set the frequency
        to the found value
        """
        closest_preset = None
        for frequency in self._presets:
            preset = self._presets[frequency]
            if self._frequency is None or abs(frequency - target_frequency) < abs(
                self._frequency - target_frequency
            ):
                self._frequency = frequency
                closest_preset = preset

        return closest_preset

    @property
    def frequency(self):
        """Return the current frequency"""
        return self._frequency

    def write(self, buf, start=0, end=None):
        """Write data from the buffer to SPI"""
        end = end if end else len(buf)
        self._transmit(buf[start:end])

    # pylint: disable=unused-argument
    def readinto(self, buf, start=0, end=None, write_value=0):
        """Read data from SPI and into the buffer"""
        end = end if end else len(buf)
        result = self._transmit([write_value] * (end - start), end - start)
        for i, b in enumerate(result):
            buf[start + i] = b

    # pylint: enable=unused-argument

    # pylint: disable=too-many-arguments
    def write_readinto(
        self, buffer_out, buffer_in, out_start=0, out_end=None, in_start=0, in_end=None
    ):
        """Perform a half-duplex write from buffer_out and then
        read data into buffer_in
        """
        out_end = out_end if out_end else len(buffer_out)
        in_end = in_end if in_end else len(buffer_in)

        result = self._transmit(buffer_out[out_start:out_end], in_end - in_start)
        for i, b in enumerate(result):
            buffer_in[in_start + i] = b

    # pylint: enable=too-many-arguments

    def _transmit(self, data, receive_length=None):
        data_to_transmit = bytearray(data)
        data_received = bytearray()

        if receive_length is None:
            receive_length = len(data)

        # If we need to receive more than we've transmitted, extend the data out.
        if receive_length > len(data):
            padding = receive_length - len(data)
            data_to_transmit.extend([0] * padding)

        # Transmit our data in chunks of the buffer size.
        while data_to_transmit:
            # Extract a single data chunk from the transmit buffer.
            chunk = data_to_transmit[0 : self.buffer_size]
            del data_to_transmit[0 : self.buffer_size]

            # Finally, exchange the data.
            response = self._spi.clock_data(len(chunk), bytes(chunk))
            data_received.extend(response)

        # Once we're done, return the data received.
        return bytes(data_received)



"""Pin definitions for the PineH64."""

from adafruit_blinka.microcontroller.allwinner.h6 import pin

D2 = pin.PD26
D3 = pin.PD25
D4 = pin.PL8
D5 = pin.PH2
D6 = pin.PG14
D7 = pin.PC16
D8 = pin.PH3
D9 = pin.PH6
D10 = pin.PH5
D11 = pin.PH4
D12 = pin.PD22
D13 = pin.PD21
D14 = pin.PD19
D15 = pin.PD20
D16 = pin.PD24
D17 = pin.PL9
D18 = pin.PG11
D19 = pin.PG10
D21 = pin.PG12
D22 = pin.PG13
D23 = pin.PD16
D24 = pin.PD17
D25 = pin.PD18
D26 = pin.PD23
D27 = pin.PD14

SDA = D2
SCL = D3

SCLK = D11
MOSI = D10
MISO = D9
CS = D8
SCK = SCLK

UART_TX = D14
UART_RX = D15



import time
import hid
import busio

from adafruit_blinka.microcontroller.mcp2221.mcp2221 import mcp2221 as _mcp2221
from adafruit_blinka.microcontroller.mcp2221.mcp2221 import MCP2221 as _MCP2221
from adafruit_blinka.microcontroller.mcp2221.i2c import I2C as _MCP2221I2C

MLXADDR = 0x33
ADDRID1 = 0x2407


class MCP2221(_MCP2221):  # pylint: disable=too-few-public-methods
    def __init__(self, address):
        self._hid = hid.device()
        self._hid.open_path(address)
        print("Connected to " + str(address))
        self._gp_config = [0x07] * 4  # "don't care" initial value
        for pin in range(4):
            self.gp_set_mode(pin, self.GP_GPIO)  # set to GPIO mode
            self.gpio_set_direction(pin, 1)  # set to INPUT


class MCP2221I2C(_MCP2221I2C):  # pylint: disable=too-few-public-methods
    def __init__(self, mcp2221, *, frequency=100000):
        self._mcp2221 = mcp2221
        self._mcp2221.i2c_configure(frequency)


class I2C(busio.I2C):  # pylint: disable=too-few-public-methods
    def __init__(self, mcp2221_i2c):
        self._i2c = mcp2221_i2c


addresses = [mcp["path"] for mcp in hid.enumerate(0x04D8, 0x00DD)]

i2c_devices = []
i2c_devices.append(I2C(MCP2221I2C(_mcp2221)))

for addr in addresses:
    try:
        i2c_device = I2C(MCP2221I2C(MCP2221(addr)))
        i2c_devices.append(i2c_device)
    except OSError:
        print("Device path: " + str(addr) + " is used")


while True:
    for i2c in i2c_devices:
        addrbuf = bytearray(2)
        addrbuf[0] = ADDRID1 >> 8
        addrbuf[1] = ADDRID1 & 0xFF
        inbuf = bytearray(6)
        i2c.writeto_then_readfrom(MLXADDR, addrbuf, inbuf)
        print("Device " + str(i2c_devices.index(i2c)) + ": ")
        print(inbuf)
        time.sleep(0.5)



"""
`micropython` - MicroPython Specific Decorator Functions
=================================================

* Author(s): cefn
"""


def const(x):
    "Emulate making a constant"
    return x


def native(f):
    "Emulate making a native"
    return f


def viper(f):
    "User is attempting to use a viper code emitter"
    raise SyntaxError("invalid micropython decorator")


def asm_thumb(f):
    "User is attempting to use an inline assembler"
    raise SyntaxError("invalid micropython decorator")



"""MIPS 24kec pin names"""
from adafruit_blinka.microcontroller.generic_linux.periphery_pin import Pin

GPIO0 = Pin(0)
GPIO1 = Pin(1)
GPIO2 = Pin(2)
GPIO3 = Pin(3)
GPIO4 = Pin(4)  # I2C SDA
GPIO5 = Pin(5)  # I2C SCL
GPIO6 = Pin(6)  # SPI CS
GPIO7 = Pin(7)  # SPI SCLK
GPIO8 = Pin(8)  # SPI MOSI
GPIO9 = Pin(9)  # SPI MISO
GPIO10 = Pin(10)

GPIO11 = Pin(11)
GPIO12 = Pin(12)
GPIO13 = Pin(13)
GPIO14 = Pin(14)
GPIO15 = Pin(15)
GPIO16 = Pin(16)
GPIO17 = Pin(17)

GPIO18 = Pin(18)
GPIO19 = Pin(19)

GPIO20 = Pin(20)
GPIO21 = Pin(21)
GPIO22 = Pin(22)
GPIO23 = Pin(23)

GPIO37 = Pin((1, 5))
GPIO38 = Pin((1, 6))
GPIO43 = Pin((1, 11))
GPIO44 = Pin((1, 12))
GPIO45 = Pin((1, 13))
GPIO46 = Pin((1, 14))

UART0_TX = GPIO12
UART0_RX = GPIO13

UART1_TX = GPIO45
UART1_RX = GPIO46

UART2_TX = GPIO20
UART2_RX = GPIO21

SPI0_MOSI = GPIO8
SPI0_MISO = GPIO9
SPI0_SCLK = GPIO7
SPI0_CS = GPIO6

I2C0_SDA = GPIO4
I2C0_SCL = GPIO5

# ordered as i2cId, sclId, sdaId
i2cPorts = ((0, I2C0_SCL, I2C0_SDA),)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = (((0, 1), SPI0_SCLK, SPI0_MOSI, SPI0_MISO),)

# ordered as uartId, txId, rxId
uartPorts = (
    (0, UART0_TX, UART0_RX),
    (1, UART1_TX, UART1_RX),
)



"""AM335x pin names"""
import Adafruit_BBIO.GPIO as GPIO


class Pin:
    """Pins dont exist in CPython so...lets make our own!"""

    IN = 0
    OUT = 1
    LOW = 0
    HIGH = 1
    PULL_NONE = 0
    PULL_UP = 1
    PULL_DOWN = 2

    id = None
    _value = LOW
    _mode = IN

    def __init__(self, pin_name):
        self.id = pin_name

    def __repr__(self):
        return str(self.id)

    def __eq__(self, other):
        return self.id == other

    def init(self, mode=IN, pull=None):
        """Initialize the Pin"""
        if mode is not None:
            if mode == self.IN:
                self._mode = self.IN
                GPIO.setup(self.id, GPIO.IN)
            elif mode == self.OUT:
                self._mode = self.OUT
                GPIO.setup(self.id, GPIO.OUT)
            else:
                raise RuntimeError("Invalid mode for pin: %s" % self.id)
        if pull is not None:
            if self._mode != self.IN:
                raise RuntimeError("Cannot set pull resistor on output")
            if pull == self.PULL_UP:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            elif pull == self.PULL_DOWN:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            else:
                raise RuntimeError("Invalid pull for pin: %s" % self.id)

    def value(self, val=None):
        """Set or return the Pin Value"""
        if val is not None:
            if val == self.LOW:
                self._value = val
                GPIO.output(self.id, val)
            elif val == self.HIGH:
                self._value = val
                GPIO.output(self.id, val)
            else:
                raise RuntimeError("Invalid value for pin")
            return None
        return GPIO.input(self.id)


# names in comments copied from
# https://github.com/adafruit/adafruit-beaglebone-io-python/blob/master/source/common.c#L73

# PocketBeagle
# P1_1 = SYS VIN        # VIN_AC
P1_2 = Pin("P1_2")  # GPIO2_23 - GPIO_87
P1_3 = Pin("P1_3")  # USB1_VBUS_OUT - (silkscreen: USB1 V_EN)
P1_4 = Pin("P1_4")  # GPIO2_25 - GPIO_89
# P1_5 = USB VBUS       # USB1_VBUS_IN
P1_6 = Pin("P1_6")  # SPI0_CS0 - GPIO_5
# P1_7 = USB VIN        # VIN-USB
P1_8 = Pin("P1_8")  # SPI0_SCLK - GPIO_2
# P1_9 = USB1 DN        # USB1-DN
P1_10 = Pin("P1_10")  # SPI0_D0 - GPIO_3
# P1_11 = USB1 DP       # USB1-DP
P1_12 = Pin("P1_12")  # SPI0_D1 - GPIO_4
# P1_13 = USB1 ID       # USB1-ID
# P1_14 = SYS 3.3V      # VOUT-3.3V
# P1_15 = SYS GND       # GND
# P1_16 = SYS GND       # GND
# P1_17 = AIN 1.8V REF- # VREFN
# P1_18 = AIN 1.8V REF+ # VREFP
P1_19 = Pin("P1_19")  # AIN0
P1_20 = Pin("P1_20")  # GPIO0_20 - GPIO_20
P1_21 = Pin("P1_21")  # AIN1
# P1_22 = SYS GND       # GND
P1_23 = Pin("P1_23")  # AIN2
# P1_22 = SYS VOUT      # VOUT-5V
P1_25 = Pin("P1_25")  # AIN3
P1_26 = Pin("P1_26")  # I2C2_SDA - GPIO_12
P1_27 = Pin("P1_27")  # AIN4
P1_28 = Pin("P1_28")  # I2C2_SCL - GPIO_13
P1_29 = Pin("P1_29")  # GPIO3_21 - GPIO_117
P1_30 = Pin("P1_30")  # UART0_TXD - GPIO_43
P1_31 = Pin("P1_31")  # GPIO3_18 - GPIO_114
P1_32 = Pin("P1_32")  # UART0_RXD - GPIO_42
P1_33 = Pin("P1_33")  # GPIO3_15 - GPIO_111
P1_34 = Pin("P1_34")  # GPIO0_26 - GPIO_26
P1_35 = Pin("P1_35")  # GPIO2_24 - GPIO_88
P1_36 = Pin("P1_36")  # EHRPWM0A - GPIO_110


P2_1 = Pin("P2_1")  # EHRPWM1A - GPIO_50
P2_2 = Pin("P2_2")  # GPIO1_27 - GPIO_59
P2_3 = Pin("P2_3")  # GPIO0_23 - GPIO_23
P2_4 = Pin("P2_4")  # GPIO1_26 - GPIO_58
P2_5 = Pin("P2_5")  # UART4_RXD - GPIO_30
P2_6 = Pin("P2_6")  # GPIO1_25 - GPIO_57
P2_7 = Pin("P2_7")  # UART4_TXD - GPIO_31
P2_8 = Pin("P2_8")  # GPIO1_28 - GPIO_60
P2_9 = Pin("P2_9")  # I2C1_SCL - GPIO_15
P2_10 = Pin("P2_10")  # GPIO1_20 - GPIO_52
P2_11 = Pin("P2_11")  # I2C1_SDA - GPIO_14
# P2_12 = SYS  PWR BTN  # POWER_BUTTON
# P2_13 = SYS VOUT      # VOUT-5V
# P2_14 = BAT VIN       # BAT-VIN
# P2_15 = SYS GND       # GND
# P2_16 = BAT TEMP      # BAT-TEMP
P2_17 = Pin("P2_17")  # GPIO2_1 - GPIO_65
P2_18 = Pin("P2_18")  # GPIO1_15 - GPIO_47
P2_19 = Pin("P2_19")  # GPIO0_27 - GPIO_27
P2_20 = Pin("P2_20")  # GPIO2_0 - GPIO_64
# P2_21 = SYS GND       # GND
P2_22 = Pin("P2_22")  # GPIO1_14 - GPIO_46
# P2_23 = SYS 3.3V      # VOUT-3.3V
P2_24 = Pin("P2_24")  # GPIO1_12 - GPIO_44
P2_25 = Pin("P2_25")  # SPI1_D1 - GPIO_41
# P2_26 = SYS NRST      # RESET#
P2_27 = Pin("P2_27")  # SPI1_D0 - GPIO_40
P2_28 = Pin("P2_28")  # GPIO3_20 - GPIO_116
P2_29 = Pin("P2_29")  # SPI1_SCLK - GPIO_7
P2_30 = Pin("P2_30")  # GPIO3_17 - GPIO_113
P2_31 = Pin("P2_31")  # SPI1_CS1 - GPIO_19
P2_32 = Pin("P2_32")  # GPIO3_16 - GPIO_112
P2_33 = Pin("P2_33")  # GPIO1_13 - GPIO_45
P2_34 = Pin("P2_34")  # GPIO3_19 - GPIO_115
P2_35 = Pin("P2_35")  # GPIO2_22 - GPIO_86
P2_36 = Pin("P2_36")  # AIN7


# BeagleBone Black
# P8_1 = DGND           # DGND - GPIO_0
# P8_2 = DGND           # DGND - GPIO_0
P8_3 = Pin("P8_3")  # GPIO1_6 - GPIO_38
P8_4 = Pin("P8_4")  # GPIO1_7 - GPIO_39
P8_5 = Pin("P8_5")  # GPIO1_2 - GPIO_34
P8_6 = Pin("P8_6")  # GPIO1_3 - GPIO_35
P8_7 = Pin("P8_7")  # TIMER4 - GPIO_66
P8_8 = Pin("P8_8")  # TIMER7 - GPIO_67
P8_9 = Pin("P8_9")  # TIMER5 - GPIO_69
P8_10 = Pin("P8_10")  # TIMER6 - GPIO_68
P8_11 = Pin("P8_11")  # GPIO1_13 - GPIO_45
P8_12 = Pin("P8_12")  # GPIO1_12 - GPIO_44
P8_13 = Pin("P8_13")  # EHRPWM2B - GPIO_23
P8_14 = Pin("P8_14")  # GPIO0_26 - GPIO_26
P8_15 = Pin("P8_15")  # GPIO1_15 - GPIO_47
P8_16 = Pin("P8_16")  # GPIO1_14 - GPIO_46
P8_17 = Pin("P8_17")  # GPIO0_27 - GPIO_27
P8_18 = Pin("P8_18")  # GPIO2_1 - GPIO_65
P8_19 = Pin("P8_19")  # EHRPWM2A - GPIO_22
P8_20 = Pin("P8_20")  # GPIO1_31 - GPIO_63
P8_21 = Pin("P8_21")  # GPIO1_30 - GPIO_62
P8_22 = Pin("P8_22")  # GPIO1_5 - GPIO_37
P8_23 = Pin("P8_23")  # GPIO1_4 - GPIO_36
P8_24 = Pin("P8_24")  # GPIO1_1 - GPIO_33
P8_25 = Pin("P8_25")  # GPIO1_0 - GPIO_32
P8_26 = Pin("P8_26")  # GPIO1_29 - GPIO_61
P8_27 = Pin("P8_27")  # GPIO2_22 - GPIO_86
P8_28 = Pin("P8_28")  # GPIO2_24 - GPIO_88
P8_29 = Pin("P8_29")  # GPIO2_23 - GPIO_87
P8_30 = Pin("P8_30")  # GPIO2_25 - GPIO_89
P8_31 = Pin("P8_31")  # UART5_CTSN - GPIO_10
P8_32 = Pin("P8_32")  # UART5_RTSN - GPIO_11
P8_33 = Pin("P8_33")  # UART4_RTSN - GPIO_9
P8_34 = Pin("P8_34")  # UART3_RTSN - GPIO_81
P8_35 = Pin("P8_35")  # UART4_CTSN - GPIO_8
P8_36 = Pin("P8_36")  # UART3_CTSN - GPIO_80
P8_37 = Pin("P8_37")  # UART5_TXD - GPIO_78
P8_38 = Pin("P8_38")  # UART5_RXD - GPIO_79
P8_39 = Pin("P8_39")  # GPIO2_12 - GPIO_76
P8_40 = Pin("P8_40")  # GPIO2_13 - GPIO_77
P8_41 = Pin("P8_41")  # GPIO2_10 - GPIO_74
P8_42 = Pin("P8_42")  # GPIO2_11 - GPIO_75
P8_43 = Pin("P8_43")  # GPIO2_8 - GPIO_72
P8_44 = Pin("P8_44")  # GPIO2_9 - GPIO_73
P8_45 = Pin("P8_45")  # GPIO2_6 - GPIO_70
P8_46 = Pin("P8_46")  # GPIO2_7 - GPIO_71

# P9_1 = DGND           # DGND - GPIO_0
# P9_2 = DGND           # DGND - GPIO_0
# P9_3 = VDD_3V3        # VDD_3V3 - GPIO_0
# P9_4 = VDD_3V3        # VDD_3V3 - GPIO_0
# P9_5 = VDD_5V         # VDD_5V - GPIO_0
# P9_6 = VDD_5V         # VDD_5V - GPIO_0
# P9_7 = SYS_5V         # SYS_5V - GPIO_0
# P9_8 = SYS_5V         # SYS_5V - GPIO_0
# P9_9 = PWR_BUT        # PWR_BUT - GPIO_0
# P9_10 = SYS_RESETN    # SYS_RESETn - GPIO_0
P9_11 = Pin("P9_11")  # UART4_RXD - GPIO_30
P9_12 = Pin("P9_12")  # GPIO1_28 - GPIO_60
P9_13 = Pin("P9_13")  # UART4_TXD - GPIO_31
P9_14 = Pin("P9_14")  # EHRPWM1A - GPIO_50
P9_15 = Pin("P9_15")  # GPIO1_16 - GPIO_48
P9_16 = Pin("P9_16")  # EHRPWM1B - GPIO_51
P9_17 = Pin("P9_17")  # I2C1_SCL - GPIO_5
P9_18 = Pin("P9_18")  # I2C1_SDA - GPIO_4
P9_19 = Pin("P9_19")  # I2C2_SCL - GPIO_13
P9_20 = Pin("P9_20")  # I2C2_SDA - GPIO_12
P9_21 = Pin("P9_21")  # UART2_TXD - GPIO_3
P9_22 = Pin("P9_22")  # UART2_RXD - GPIO_2
P9_23 = Pin("P9_23")  # GPIO1_17 - GPIO_49
P9_24 = Pin("P9_24")  # UART1_TXD - GPIO_15
P9_25 = Pin("P9_25")  # GPIO3_21 - GPIO_117
P9_26 = Pin("P9_26")  # UART1_RXD - GPIO_14
P9_27 = Pin("P9_27")  # GPIO3_19 - GPIO_115
P9_28 = Pin("P9_28")  # SPI1_CS0 - GPIO_113
P9_29 = Pin("P9_29")  # SPI1_D0 - GPIO_111
P9_30 = Pin("P9_30")  # SPI1_D1 - GPIO_112
P9_31 = Pin("P9_31")  # SPI1_SCLK - GPIO_110
# P9_32 = VDD_ADC       # VDD_ADC - GPIO_0
# P9_33 = AIN4          # AIN4 - GPIO_0
# P9_34 = GNDA_ADC      # GNDA_ADC - GPIO_0
# P9_35 = AIN6          # AIN6 - GPIO_0
# P9_36 = AIN5          # AIN5 - GPIO_0
# P9_37 = AIN2          # AIN2 - GPIO_0
# P9_38 = AIN3          # AIN3 - GPIO_0
# P9_39 = AIN0          # AIN0 - GPIO_0
# P9_40 = AIN1          # AIN1 - GPIO_0
P9_41 = Pin("P9_41")  # CLKOUT2 - GPIO_20
P9_42 = Pin("P9_42")  # GPIO0_7 - GPIO_7
# P9_43 = DGND          # DGND - GPIO_0
# P9_44 = DGND          # DGND - GPIO_0
# P9_45 = DGND          # DGND - GPIO_0
# P9_46 = DGND          # DGND - GPIO_0


##########################################
# common to all beagles
USR0 = Pin("USR0")  # USR0 - GPIO_53
USR1 = Pin("USR1")  # USR1 - GPIO_54
USR2 = Pin("USR2")  # USR2 - GPIO_55
USR3 = Pin("USR3")  # USR3 - GPIO_56


##########################################
# specials

# analog input
AIN0 = Pin("AIN0")
AIN1 = Pin("AIN1")
AIN2 = Pin("AIN2")
AIN3 = Pin("AIN3")
AIN4 = Pin("AIN4")
AIN5 = Pin("AIN5")
AIN6 = Pin("AIN6")
AIN7 = Pin("AIN7")

# PWM
EHRPWM0A = Pin("EHRPWM0A")
EHRPWM0B = Pin("EHRPWM0B")
EHRPWM1A = Pin("EHRPWM1A")
EHRPWM1B = Pin("EHRPWM1B")
EHRPWM2A = Pin("EHRPWM2A")
EHRPWM2B = Pin("EHRPWM2B")
ECAPPWM0 = Pin("ECAPPWM0")
ECAPPWM2 = Pin("ECAPPWM2")
TIMER4 = Pin("TIMER4")
TIMER5 = Pin("TIMER5")
TIMER6 = Pin("TIMER6")
TIMER7 = Pin("TIMER7")


# I2C1
I2C1_SDA = Pin("I2C1_SDA")
I2C1_SCL = Pin("I2C1_SCL")

# I2C2
I2C2_SDA = Pin("I2C2_SDA")
I2C2_SCL = Pin("I2C2_SCL")

# SPI0
SPI0_CS0 = Pin("SPI0_CS0")
SPI0_SCLK = Pin("SPI0_SCLK")
SPI0_D1 = Pin("SPI0_D1")
SPI0_D0 = Pin("SPI0_D0")

# SPI1
SPI1_CS0 = Pin("SPI1_CS0")
SPI1_CS1 = Pin("SPI1_CS1")
SPI1_SCLK = Pin("SPI1_SCLK")
SPI1_D1 = Pin("SPI1_D1")
SPI1_D0 = Pin("SPI1_D0")

# UART0
UART0_TXD = Pin("UART0_TXD")
UART0_RXD = Pin("UART0_RXD")

# UART1
UART1_TXD = Pin("UART1_TXD")
UART1_RXD = Pin("UART1_RXD")
UART1_RTSn = Pin("UART1_RTSn")
UART1_CTSn = Pin("UART1_CTSn")

# UART2
UART2_TXD = Pin("UART2_TXD")
UART2_RXD = Pin("UART2_RXD")

# UART3
UART3_TXD = Pin("UART3_TXD")
UART3_RXD = Pin("UART3_RXD")
UART3_RTSn = Pin("UART3_RTSn")
UART3_CTSn = Pin("UART3_CTSn")

# UART4
UART4_TXD = Pin("UART4_TXD")
UART4_RXD = Pin("UART4_RXD")
UART4_RTSn = Pin("UART4_RTSn")
UART4_CTSn = Pin("UART4_CTSn")

# UART5
UART5_TXD = Pin("UART5_TXD")
UART5_RXD = Pin("UART5_RXD")
UART5_RTSn = Pin("UART5_RTSn")
UART5_CTSn = Pin("UART5_CTSn")


# ordered as spiId, sckId, mosiId, misoId
spiPorts = (
    (0, SPI0_SCLK, SPI0_D1, SPI0_D0),
    (1, SPI1_SCLK, SPI1_D1, SPI1_D0),
)

# ordered as uartId, txId, rxId
uartPorts = (
    # (0, UART0_TXD, UART0_RXD),
    # (1, UART1_TXD, UART1_RXD),
    # (2, UART2_TXD, UART2_RXD),
    # (4, UART4_TXD, UART4_RXD),
    # (5, UART5_TXD, UART5_RXD),
)

# ordered as i2cId, SCL, SDA
i2cPorts = (
    (1, I2C1_SCL, I2C1_SDA),
    (2, I2C2_SCL, I2C2_SDA),
)

PWM1 = P1_36
PWM2 = P1_33
PWM3 = P2_1
PWM4 = P2_3

pwmOuts = (((0, 0), PWM1), ((0, 1), PWM2), ((2, 0), PWM3), ((4, 1), PWM4))



import unittest


class TestEnum(unittest.TestCase):
    """
    Verifies the repl() and str() behaviour of an example Enum
    Enums represent configuration values such as digitalio.Direction, digitalio.Pull etc.
    """

    def setUp(self):
        """Create an example Enum, mocking __module__ and __qualname__"""
        import adafruit_blinka

        class Cls(adafruit_blinka.Enum):
            pass

        Cls.one = Cls()
        Cls.two = Cls()
        # class refs would be implicitly populated correctly in a real module
        Cls.__module__ = "ho.hum"
        Cls.__qualname__ = "Example"
        self.Cls = Cls

    def test_iteritems(self):
        """A subtype of Enum can list all attributes of its own type"""
        items = list(self.Cls.iteritems())
        self.assertEqual(
            items,
            [
                ("one", self.Cls.one),
                ("two", self.Cls.two),
            ],
        )

    def test_repr(self):
        """A repr() call on an Enum gives its fully-qualified name"""
        name = "one"
        actual = repr(getattr(self.Cls, name))
        expected = "{}.{}.{}".format(self.Cls.__module__, self.Cls.__qualname__, name)
        self.assertEqual(actual, expected)

    def test_str(self):
        """A str() call on an Enum performs the same as repr()"""
        self.assertEqual(str(self.Cls.one), repr(self.Cls.one))


class TestDigitalInOut(unittest.TestCase):
    def test_context_manager(self):
        import digitalio
        from testing.board import default_pin

        """Deinitialisation is triggered by __exit__() and should dispose machine.pin reference"""
        dio = digitalio.DigitalInOut(default_pin)
        self.assertIsNotNone(dio._pin)
        with dio:
            pass
        self.assertIsNone(dio._pin)



import sys
import time
import adafruit_blinka.agnostic as agnostic
import board
import digitalio

# from Adafruit_GPIO import Platform
# print("Platform = ", Platform.platform_detect(), Platform.pi_version())

print("hello blinka!")

print(
    "Found system type: %s (sys.platform %s implementation %s) "
    % (agnostic.board_id, sys.platform, sys.implementation.name)
)

print("board contents: ", dir(board))


led = digitalio.DigitalInOut(board.D4)
led.direction = digitalio.Direction.OUTPUT

button = digitalio.DigitalInOut(board.D18)
button.direction = digitalio.Direction.INPUT
button.pull = digitalio.Pull.DOWN

while True:
    led.value = button.value
    time.sleep(0.1)



"""Allwinner H3 pin names"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

PA0 = Pin(0)
UART2_TX = PA0
PA1 = Pin(1)
UART2_RX = PA1
PA2 = Pin(2)
PA3 = Pin(3)
PA6 = Pin(6)
PA7 = Pin(7)
PA8 = Pin(8)
PA9 = Pin(9)
PA10 = Pin(10)
PA11 = Pin(11)
TWI0_SCL = PA11
PA12 = Pin(12)
TWI0_SDA = PA12
PA13 = Pin(13)
UART3_TX = PA13
PA14 = Pin(14)
UART3_RX = PA14
SPI1_SCLK = PA14
PA15 = Pin(15)
SPI1_MOSI = PA15
PA16 = Pin(16)
SPI1_MISO = PA16
PA17 = Pin(17)
PA18 = Pin(18)
PA19 = Pin(19)
PA20 = Pin(20)
PA21 = Pin(21)

PC0 = Pin(64)
SPI0_MOSI = PC0
PC1 = Pin(65)
SPI0_MISO = PC1
PC2 = Pin(66)
SPI0_SCLK = PC2
PC3 = Pin(67)
SPI0_CS = PC3
PC4 = Pin(68)
PC7 = Pin(71)

PD14 = Pin(110)

PG6 = Pin(198)
UART1_TX = PG6
PG7 = Pin(199)
UART1_RX = PG7
PG8 = Pin(200)
PG9 = Pin(201)
PG10 = Pin(202)
PG11 = Pin(203)
PG12 = Pin(204)
PG13 = Pin(205)

PL2 = Pin((1, 2))
PL4 = Pin((1, 4))

i2cPorts = ((0, TWI0_SCL, TWI0_SDA),)
# ordered as spiId, sckId, mosiId, misoId
spiPorts = (
    (0, SPI0_SCLK, SPI0_MOSI, SPI0_MISO),
    (1, SPI1_SCLK, SPI1_MOSI, SPI1_MISO),
)
# ordered as uartId, txId, rxId
uartPorts = ((3, UART3_TX, UART3_RX),)



"""Pin definitions for the Hifive Unleashed."""

from adafruit_blinka.microcontroller.hfu540 import pin

GPIO_A = pin.GPIO0
GPIO_B = pin.GPIO1
GPIO_C = pin.GPIO2
GPIO_D = pin.GPIO3
GPIO_E = pin.GPIO4
GPIO_F = pin.GPIO5
GPIO_G = pin.GPIO6
GPIO_H = pin.GPIO7
GPIO_I = pin.GPIO8
GPIO_J = pin.GPIO9
GPIO_K = pin.GPIO15

UART0_TXD = pin.UART0_TXD
UART0_RXD = pin.UART0_RXD
SPI0_SCLK = pin.SPI0_SCLK
SPI0_DIN = pin.SPI0_DIN
UART1_TXD = pin.UART1_TXD
SPI0_CS = pin.SPI0_CS
UART1_RXD = pin.UART1_RXD
SPI0_DOUT = pin.SPI0_DOUT
I2C0_SCL = pin.I2C0_SCL
I2C0_SDA = pin.I2C0_SDA

SDA = pin.I2C0_SDA
SCL = pin.I2C0_SCL

I2C0_SDA = pin.I2C0_SDA
I2C0_SCL = pin.I2C0_SCL

SCLK = pin.SPI0_SCLK
MOSI = pin.SPI0_DOUT
MISO = pin.SPI0_DIN
SPI_CS = pin.SPI0_CS



"""Pin definitions for the Onion Omega2+ Expansion Dock."""

from adafruit_blinka.microcontroller.mips24kec import pin

D0 = pin.GPIO0
D1 = pin.GPIO1
D2 = pin.GPIO2
D3 = pin.GPIO3

D4 = pin.GPIO4
D5 = pin.GPIO5
D6 = pin.GPIO6
D7 = pin.GPIO7
D8 = pin.GPIO8
D9 = pin.GPIO9

D11 = pin.GPIO11
D15 = pin.GPIO15  # RGB LED Blue
D16 = pin.GPIO16  # RGB LED Green
D17 = pin.GPIO17  # RGB LED Red
D18 = pin.GPIO18
D19 = pin.GPIO19

SDA = pin.I2C0_SDA
SCL = pin.I2C0_SCL

SCLK = pin.SPI0_SCLK
MOSI = pin.SPI0_MOSI
MISO = pin.SPI0_MISO
CS = pin.SPI0_CS
SCK = SCLK

UART_TX = pin.UART1_TX
UART_RX = pin.UART1_RX



"""Pin definitions for the Orange Pi PC."""

from adafruit_blinka.microcontroller.allwinner.h3 import pin

PA12 = pin.PA12
SDA = pin.PA12
PA11 = pin.PA11
SCL = pin.PA11
PA6 = pin.PA6
PA1 = pin.PA1
PA0 = pin.PA0
PA3 = pin.PA3
PC0 = pin.PC0
MOSI = pin.PC0
MOSI1 = pin.SPI1_MOSI
PC1 = pin.PC1
MISO = pin.PC1
PC2 = pin.PC2
SCK = pin.PC2
SCK1 = pin.SPI1_SCLK
SCLK = pin.PC2
PA19 = pin.PA19
PA7 = pin.PA7
PA8 = pin.PA8
PA9 = pin.PA9
PA10 = pin.PA10
PA20 = pin.PA20

PA13 = pin.PA13
TX = pin.PA13
PA14 = pin.PA14
RX = pin.PA14
PD14 = pin.PD14
PC4 = pin.PC4
PC7 = pin.PC7
PA2 = pin.PA2
PC3 = pin.PC3
PA21 = pin.PA21
PA18 = pin.PA18
PG8 = pin.PG8
PG9 = pin.PG9
PG6 = pin.PG6
PG7 = pin.PG7



"""Microcontroller pins"""

from adafruit_platformdetect.constants import chips as ap_chip
from adafruit_blinka import Enum
from adafruit_blinka.agnostic import board_id, chip_id


class Pin(Enum):
    """Reference Pin object"""

    def __init__(self, pin_id):
        """Identifier for pin, referencing platform-specific pin id"""
        self._id = pin_id

    def __repr__(self):
        # pylint: disable=import-outside-toplevel
        import board

        for key in dir(board):
            if getattr(board, key) is self:
                return "board.{}".format(key)
        import microcontroller.pin as pin

        # pylint: enable=import-outside-toplevel

        for key in dir(pin):
            if getattr(pin, key) is self:
                return "microcontroller.pin.{}".format(key)
        return repr(self)


# We intentionally are patching into this namespace so skip the wildcard check.
# pylint: disable=unused-wildcard-import,wildcard-import,ungrouped-imports

if chip_id == ap_chip.ESP8266:
    from adafruit_blinka.microcontroller.esp8266 import *
elif chip_id == ap_chip.STM32F405:
    from adafruit_blinka.microcontroller.stm32.stm32f405 import *
elif chip_id == ap_chip.BCM2XXX:
    from adafruit_blinka.microcontroller.bcm283x import *
elif chip_id == ap_chip.DRA74X:
    from adafruit_blinka.microcontroller.dra74x.pin import *
elif chip_id == ap_chip.AM33XX:
    from adafruit_blinka.microcontroller.am335x import *
elif chip_id == ap_chip.SUN8I:
    from adafruit_blinka.microcontroller.allwinner.h3 import *
elif chip_id == ap_chip.SAMA5:
    from adafruit_blinka.microcontroller.sama5 import *
elif chip_id == ap_chip.T210:
    from adafruit_blinka.microcontroller.tegra.t210 import *
elif chip_id == ap_chip.T186:
    from adafruit_blinka.microcontroller.tegra.t186 import *
elif chip_id == ap_chip.T194:
    from adafruit_blinka.microcontroller.tegra.t194 import *
elif chip_id == ap_chip.S905:
    from adafruit_blinka.microcontroller.amlogic.s905.pin import *
elif chip_id == ap_chip.S905X3:
    from adafruit_blinka.microcontroller.amlogic.s905x3.pin import *
elif chip_id == ap_chip.S922X:
    from adafruit_blinka.microcontroller.amlogic.s922x.pin import *
elif chip_id == ap_chip.EXYNOS5422:
    from adafruit_blinka.microcontroller.samsung.exynos5422.pin import *
elif chip_id == ap_chip.APQ8016:
    from adafruit_blinka.microcontroller.snapdragon.apq8016.pin import *
elif chip_id == ap_chip.A64:
    from adafruit_blinka.microcontroller.allwinner.a64.pin import *
elif chip_id == ap_chip.A33:
    from adafruit_blinka.microcontroller.allwinner.a33.pin import *
elif chip_id == ap_chip.RK3308:
    from adafruit_blinka.microcontroller.rockchip.rk3308.pin import *
elif chip_id == ap_chip.H5:
    from adafruit_blinka.microcontroller.allwinner.h5.pin import *
elif chip_id == ap_chip.IMX8MX:
    from adafruit_blinka.microcontroller.nxp_imx8m import *
elif chip_id == ap_chip.IMX6ULL:
    from adafruit_blinka.microcontroller.nxp_imx6ull import *
elif chip_id == ap_chip.HFU540:
    from adafruit_blinka.microcontroller.hfu540.pin import *
elif chip_id == ap_chip.BINHO:
    from adafruit_blinka.microcontroller.nova import *
elif chip_id == ap_chip.LPC4330:
    from adafruit_blinka.microcontroller.nxp_lpc4330 import *
elif chip_id == ap_chip.MIPS24KC:
    from adafruit_blinka.microcontroller.atheros.ar9331.pin import *
elif chip_id == ap_chip.MIPS24KEC:
    from adafruit_blinka.microcontroller.mips24kec.pin import *
elif chip_id == ap_chip.PENTIUM_N3710:
    from adafruit_blinka.microcontroller.pentium.n3710.pin import *
elif chip_id == ap_chip.STM32MP157:
    from adafruit_blinka.microcontroller.stm32.stm32mp157.pin import *
elif chip_id == ap_chip.MT8167:
    from adafruit_blinka.microcontroller.mt8167.pin import *
else:
    raise NotImplementedError("Microcontroller not supported:", chip_id)



"""Custom PulseIn Class to read PWM signals"""
import time
import subprocess
import os
import atexit
import random
import sysv_ipc

DEBUG = False
queues = []
procs = []

# The message queues live outside of python space, and must be formally cleaned!
def final():
    """In case the program is cancelled or quit, we need to clean up the PulseIn
    helper process and also the message queue, this is called at exit to do so"""
    if DEBUG:
        print("Cleaning up message queues", queues)
        print("Cleaning up processes", procs)
    for q in queues:
        q.remove()
    for proc in procs:
        proc.terminate()


atexit.register(final)

# pylint: disable=c-extension-no-member
class PulseIn:
    """PulseIn Class to read PWM signals"""

    def __init__(self, pin, maxlen=2, idle_state=False):
        """Create a PulseIn object associated with the given pin.
        The object acts as a read-only sequence of pulse lengths with
        a given max length. When it is active, new pulse lengths are
        added to the end of the list. When there is no more room
        (len() == maxlen) the oldest pulse length is removed to make room."""
        self._pin = pin
        self._maxlen = maxlen
        self._idle_state = idle_state
        self._queue_key = random.randint(1, 9999)
        try:
            self._mq = sysv_ipc.MessageQueue(None, flags=sysv_ipc.IPC_CREX)
            if DEBUG:
                print("Message Queue Key: ", self._mq.key)
            queues.append(self._mq)
        except sysv_ipc.ExistentialError:
            raise RuntimeError(
                "Message queue creation failed"
            ) from sysv_ipc.ExistentialError

        dir_path = os.path.dirname(os.path.realpath(__file__))
        cmd = [
            dir_path + "/libgpiod_pulsein",
            "--pulses",
            str(maxlen),
            "--queue",
            str(self._mq.key),
        ]
        if idle_state:
            cmd.append("-i")
        cmd.append("gpiochip0")
        cmd.append(str(pin))
        if DEBUG:
            print(cmd)

        self._process = subprocess.Popen(cmd)
        procs.append(self._process)

        # wait for it to start up
        if DEBUG:
            print("Waiting for startup success message from subprocess")
        message = self._wait_receive_msg(timeout=0.25)
        if message[0] != b"!":
            raise RuntimeError("Could not establish message queue with subprocess")
        self._paused = False

    # pylint: disable=redefined-builtin
    def _wait_receive_msg(self, timeout=0, type=2):
        """Internal helper that will wait for new messages of a given type,
        and throw an exception on timeout"""
        if timeout > 0:
            stamp = time.monotonic()
            while (time.monotonic() - stamp) < timeout:
                try:
                    message = self._mq.receive(block=False, type=type)
                    return message
                except sysv_ipc.BusyError:
                    time.sleep(0.001)  # wait a bit then retry!
            # uh-oh timed out
            raise RuntimeError(
                "Timed out waiting for PulseIn message. Make sure libgpiod is installed."
            )
        message = self._mq.receive(block=True, type=type)
        return message

    # pylint: enable=redefined-builtin

    def deinit(self):
        """Deinitialises the PulseIn and releases any hardware and software
        resources for reuse."""
        # Clean up after ourselves
        self._process.terminate()
        procs.remove(self._process)
        self._mq.remove()
        queues.remove(self._mq)

    def __enter__(self):
        """No-op used by Context Managers."""
        return self

    def __exit__(self, exc_type, exc_value, tb):
        """Automatically deinitializes the hardware when exiting a context."""
        self.deinit()

    def resume(self, trigger_duration=0):
        """Resumes pulse capture after an optional trigger pulse."""
        if trigger_duration != 0:
            self._mq.send("t%d" % trigger_duration, True, type=1)
        else:
            self._mq.send("r", True, type=1)
        self._paused = False

    def pause(self):
        """Pause pulse capture"""
        self._mq.send("p", True, type=1)
        self._paused = True

    @property
    def paused(self):
        """True when pulse capture is paused as a result of pause() or
        an error during capture such as a signal that is too fast."""
        return self._paused

    @property
    def maxlen(self):
        """The maximum length of the PulseIn. When len() is equal to maxlen,
        it is unclear which pulses are active and which are idle."""
        return self._maxlen

    def clear(self):
        """Clears all captured pulses"""
        self._mq.send("c", True, type=1)

    def popleft(self):
        """Removes and returns the oldest read pulse."""
        self._mq.send("^", True, type=1)
        message = self._wait_receive_msg()
        reply = int(message[0].decode("utf-8"))
        # print(reply)
        if reply == -1:
            raise IndexError("pop from empty list")
        return reply

    def __len__(self):
        """Returns the current pulse length"""
        self._mq.send("l", True, type=1)
        message = self._wait_receive_msg()
        return int(message[0].decode("utf-8"))

    # pylint: disable=redefined-builtin
    def __getitem__(self, index, type=None):
        """Returns the value at the given index or values in slice."""
        self._mq.send("i%d" % index, True, type=1)
        message = self._wait_receive_msg()
        ret = int(message[0].decode("utf-8"))
        if ret == -1:
            raise IndexError("list index out of range")
        return ret

    # pylint: enable=redefined-builtin



import gc
from testing import yes_no

gc.collect()
from unittest import TestCase

gc.collect()
from testing.board.i2c import I2C

gc.collect()


class TestBME280Interactive(TestCase):
    def test_read_value(self):

        import board

        gc.collect()
        import adafruit_bme280

        gc.collect()

        if not (
            yes_no("Is BME280 wired to SCL {} SDA {}".format(board.SCL, board.SDA))
        ):
            return  # test trivially passed

        i2c = I2C(board.SCL, board.SDA)
        bme280 = adafruit_bme280.Adafruit_BME280_I2C(i2c)
        temperature = bme280.temperature
        humidity = bme280.humidity
        pressure = bme280.pressure
        altitude = bme280.altitude
        self.assertTrue(type(temperature) is float)
        self.assertTrue(type(humidity) is float)
        self.assertTrue(type(pressure) is float)
        self.assertTrue(type(altitude) is float)

        self.assertTrue(-50 <= temperature <= 50)
        self.assertTrue(0 <= humidity <= 100)
        self.assertTrue(900 <= pressure <= 1100)
        self.assertTrue(-1000 <= altitude <= 9, 848)


class TestMMA8451Interactive(TestCase):
    def test_read_value(self):
        import math

        gc.collect()
        import board

        gc.collect()

        if not (
            yes_no(
                "Is MMA8451 wired to SCL {} SDA {} and held still".format(
                    board.SCL, board.SDA
                )
            )
        ):
            return  # test trivially passed
        # from https://github.com/adafruit/Adafruit_CircuitPython_MMA8451/blob/29e31a0bb836367bc73763b83513105252b7b264/examples/simpletest.py
        import adafruit_mma8451

        i2c = I2C(board.SCL, board.SDA)
        sensor = adafruit_mma8451.MMA8451(i2c)

        x, y, z = sensor.acceleration
        absolute = math.sqrt(x ** 2 + y ** 2 + z ** 2)
        self.assertTrue(9 <= absolute <= 11, "Not earth gravity")

        orientation = sensor.orientation
        self.assertTrue(
            orientation
            in (
                adafruit_mma8451.PL_PUF,
                adafruit_mma8451.PL_PUB,
                adafruit_mma8451.PL_PDF,
                adafruit_mma8451.PL_PDB,
                adafruit_mma8451.PL_LRF,
                adafruit_mma8451.PL_LRB,
                adafruit_mma8451.PL_LLF,
                adafruit_mma8451.PL_LLB,
            )
        )


class TestBNO055Interactive(TestCase):
    def test_read_value(self):
        """
        Access all sensor values as per
        https://github.com/adafruit/Adafruit_CircuitPython_BNO055/blob/bdf6ada21e7552c242bc470d4d2619b480b4c574/examples/values.py
        Note I have not successfully run this test. Possibly a hardware issue with module I have.
        See https://forums.adafruit.com/viewtopic.php?f=60&t=131665
        """
        import board

        gc.collect()
        import adafruit_bno055

        gc.collect()
        i2c = I2C(board.SCL, board.SDA)
        sensor = adafruit_bno055.BNO055(i2c)

        self.assertTrue(9 <= sensor.gravity <= 11)
        self.assertTrue(sensor.temperature != 0)
        self.assertTrue(sensor.acceleration != (0, 0, 0))
        self.assertTrue(sensor.magnetometer != (0, 0, 0))
        self.assertTrue(sensor.gyroscope != (0, 0, 0))
        self.assertTrue(sensor.quaternion != (0, 0, 0, 0))
        sensor.euler
        sensor.linear_acceleration



"""Atheros AR9331 pin names"""
from adafruit_blinka.microcontroller.generic_linux.libgpiod_pin import Pin

GPIO_0 = Pin((0, 0))
GPIO_1 = Pin((0, 1))
GPIO_2 = Pin((0, 2))
GPIO_3 = Pin((0, 3))
SPI_CLK = GPIO_3
GPIO_4 = Pin((0, 4))
SPI_MOSI = GPIO_4
GPIO_5 = Pin((0, 5))
SPI_MISO = GPIO_5
GPIO_6 = Pin((0, 6))
GPIO_7 = Pin((0, 7))
GPIO_8 = Pin((0, 8))
GPIO_9 = Pin((0, 9))
UART0_RX = GPIO_9
GPIO_10 = Pin((0, 10))
UART0_TX = GPIO_10
GPIO_11 = Pin((0, 11))
GPIO_12 = Pin((0, 12))
GPIO_13 = Pin((0, 13))
GPIO_14 = Pin((0, 14))
GPIO_15 = Pin((0, 15))
GPIO_16 = Pin((0, 16))
GPIO_17 = Pin((0, 17))
GPIO_18 = Pin((0, 18))
GPIO_19 = Pin((0, 19))
GPIO_20 = Pin((0, 20))
TWI0_SCL = GPIO_20

GPIO_21 = Pin((0, 21))
TWI0_SDA = GPIO_21
GPIO_22 = Pin((0, 22))
GPIO_23 = Pin((0, 23))
GPIO_24 = Pin((0, 24))
GPIO_25 = Pin((0, 25))
GPIO_26 = Pin((0, 26))
GPIO_27 = Pin((0, 27))
GPIO_28 = Pin((0, 28))

# ordered as i2cId, sclId, sdaId
i2cPorts = (0, TWI0_SCL, TWI0_SDA)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((1, SPI_CLK, SPI_MOSI, SPI_MISO),)

# ordered as uartId, txId, rxId
uartPorts = (0, UART0_TX, UART0_RX)



"""PWMOut Class for Binho Nova"""

try:
    from microcontroller.pin import pwmOuts
except ImportError:
    raise RuntimeError("No PWM outputs defined for this board") from ImportError

from microcontroller.pin import Pin


# pylint: disable=unnecessary-pass
class PWMError(IOError):
    """Base class for PWM errors."""

    pass


# pylint: enable=unnecessary-pass


class PWMOut:
    """Pulse Width Modulation Output Class"""

    # Nova instance
    _nova = None
    MAX_CYCLE_LEVEL = 1024

    def __init__(self, pin, *, frequency=750, duty_cycle=0, variable_frequency=False):
        """Instantiate a PWM object and open the sysfs PWM corresponding to the
        specified channel and pin.

        Args:
            pin (Pin): CircuitPython Pin object to output to
            duty_cycle (int) : The fraction of each pulse which is high. 16-bit
            frequency (int) : target frequency in Hertz (32-bit)
            variable_frequency (bool) : True if the frequency will change over time

        Returns:
            PWMOut: PWMOut object.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if `channel` or `pin` types are invalid.
            ValueError: if PWM channel does not exist.

        """
        if PWMOut._nova is None:
            # pylint: disable=import-outside-toplevel
            from adafruit_blinka.microcontroller.nova import Connection

            # pylint: enable=import-outside-toplevel

            PWMOut._nova = Connection.getInstance()

        PWMOut._nova.setOperationMode(0, "IO")
        self._pwmpin = None
        self._channel = None
        self._enable = False
        self._open(pin, duty_cycle, frequency, variable_frequency)

    def __del__(self):
        self.deinit()
        PWMOut._nova.close()

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self.deinit()

    def _open(self, pin, duty=0, freq=750, variable_frequency=False):
        self._channel = None
        for pwmpair in pwmOuts:
            if pwmpair[1] == pin:
                self._channel = pwmpair[0][0]
                self._pwmpin = pwmpair[0][1]

        self._pin = pin
        if self._channel is None:
            raise RuntimeError("No PWM channel found for this Pin")

        if variable_frequency:
            print("Variable Frequency is not supported, continuing without it...")

        PWMOut._nova.setIOpinMode(self._pwmpin, Pin.PWM)

        # set frequency
        self.frequency = freq
        # set period
        self._period = self._get_period()

        # set duty
        self.duty_cycle = duty

        self._set_enabled(True)

    def deinit(self):
        """Deinit the Nova PWM."""
        # pylint: disable=broad-except
        try:
            if self._channel is not None:
                # self.duty_cycle = 0
                self._set_enabled(False)  # make to disable before unexport

        except Exception as e:
            # due to a race condition for which I have not yet been
            # able to find the root cause, deinit() often fails
            # but it does not effect future usage of the pwm pin
            print(
                "warning: failed to deinitialize pwm pin {0}:{1} due to: {2}\n".format(
                    self._channel, self._pwmpin, type(e).__name__
                )
            )
        finally:
            self._channel = None
            self._pwmpin = None
        # pylint: enable=broad-except

    def _is_deinited(self):
        if self._pwmpin is None:
            raise ValueError(
                "Object has been deinitialize and can no longer "
                "be used. Create a new object."
            )

    # Mutable properties

    def _get_period(self):
        return 1.0 / self._get_frequency()

    def _set_period(self, period):
        if not isinstance(period, (int, float)):
            raise TypeError("Invalid period type, should be int or float.")

        self._set_frequency(1.0 / period)

    period = property(_get_period, _set_period)

    """Get or set the PWM's output period in seconds.

    Raises:
        PWMError: if an I/O or OS error occurs.
        TypeError: if value type is not int or float.

    :type: int, float
    """

    def _get_duty_cycle(self):
        duty_cycle = Pin._nova.getIOpinValue(self._pwmpin)

        # Convert duty cycle to ratio from 0.0 to 1.0
        duty_cycle = duty_cycle / PWMOut.MAX_CYCLE_LEVEL

        # convert to 16-bit
        duty_cycle = int(duty_cycle * 65535)
        return duty_cycle

    def _set_duty_cycle(self, duty_cycle):
        if not isinstance(duty_cycle, (int, float)):
            raise TypeError("Invalid duty cycle type, should be int or float.")

        # convert from 16-bit
        duty_cycle /= 65535.0
        if not 0.0 <= duty_cycle <= 1.0:
            raise ValueError("Invalid duty cycle value, should be between 0.0 and 1.0.")

        # Convert duty cycle from ratio to 1024 levels
        duty_cycle = duty_cycle * PWMOut.MAX_CYCLE_LEVEL

        # Set duty cycle
        # pylint: disable=protected-access
        Pin._nova.setIOpinValue(self._pwmpin, duty_cycle)
        # pylint: enable=protected-access

    duty_cycle = property(_get_duty_cycle, _set_duty_cycle)
    """Get or set the PWM's output duty cycle as a ratio from 0.0 to 1.0.

    Raises:
        PWMError: if an I/O or OS error occurs.
        TypeError: if value type is not int or float.
        ValueError: if value is out of bounds of 0.0 to 1.0.

    :type: int, float
    """

    def _get_frequency(self):
        return int(PWMOut._nova.getIOpinPWMFreq(self._pwmpin).split("PWMFREQ ")[1])

    def _set_frequency(self, frequency):
        if not isinstance(frequency, (int, float)):
            raise TypeError("Invalid frequency type, should be int or float.")

        PWMOut._nova.setIOpinPWMFreq(self._pwmpin, frequency)

    frequency = property(_get_frequency, _set_frequency)
    """Get or set the PWM's output frequency in Hertz.

    Raises:
        PWMError: if an I/O or OS error occurs.
        TypeError: if value type is not int or float.

    :type: int, float
    """

    def _get_enabled(self):
        enabled = self._enable

        if enabled == "1":
            return True
        if enabled == "0":
            return False

        raise PWMError(None, 'Unknown enabled value: "%s"' % enabled)

    def _set_enabled(self, value):
        """Get or set the PWM's output enabled state.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not bool.

        :type: bool
        """
        if not isinstance(value, bool):
            raise TypeError("Invalid enabled type, should be string.")
        self._enable = value
        if not self._enable:
            self._set_duty_cycle(0.0)

    # String representation

    def __str__(self):
        return "PWM%d, pin %s (freq=%f Hz, duty_cycle=%f%%)" % (
            self._pin,
            self._pin,
            self.frequency,
            self.duty_cycle * 100,
        )



"""
Much code from https://github.com/vsergeev/python-periphery/blob/master/periphery/pwm.py
Copyright (c) 2015-2016 vsergeev / Ivan (Vanya) A. Sergeev
License: MIT
"""

import os

try:
    from microcontroller.pin import pwmOuts
except ImportError:
    raise RuntimeError("No PWM outputs defined for this board") from ImportError

# pylint: disable=unnecessary-pass
class PWMError(IOError):
    """Base class for PWM errors."""

    pass


# pylint: enable=unnecessary-pass


class PWMOut:
    """Pulse Width Modulation Output Class"""

    # Sysfs paths
    _sysfs_path = "/sys/class/pwm/"
    _channel_path = "pwmchip{}"

    # Channel paths
    _export_path = "export"
    _unexport_path = "unexport"
    _pin_path = "pwm-{}:{}"

    # Pin attribute paths
    _pin_period_path = "period"
    _pin_duty_cycle_path = "duty_cycle"
    _pin_polarity_path = "polarity"
    _pin_enable_path = "enable"

    def __init__(self, pin, *, frequency=500, duty_cycle=0, variable_frequency=False):
        """Instantiate a PWM object and open the sysfs PWM corresponding to the
        specified channel and pin.

        Args:
            pin (Pin): CircuitPython Pin object to output to
            duty_cycle (int) : The fraction of each pulse which is high. 16-bit
            frequency (int) : target frequency in Hertz (32-bit)
            variable_frequency (bool) : True if the frequency will change over time

        Returns:
            PWMOut: PWMOut object.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if `channel` or `pin` types are invalid.
            ValueError: if PWM channel does not exist.

        """

        self._pwmpin = None
        self._channel = None
        self._period = 0
        self._open(pin, duty_cycle, frequency, variable_frequency)

    def __del__(self):
        self.deinit()

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self.deinit()

    def _open(self, pin, duty=0, freq=500, variable_frequency=False):
        self._channel = None
        for pwmpair in pwmOuts:
            if pwmpair[1] == pin:
                self._channel = pwmpair[0][0]
                self._pwmpin = pwmpair[0][1]

        self._pin = pin
        if self._channel is None:
            raise RuntimeError("No PWM channel found for this Pin")

        if variable_frequency:
            print("Variable Frequency is not supported, continuing without it...")

        channel_path = os.path.join(
            self._sysfs_path, self._channel_path.format(self._channel)
        )
        if not os.path.isdir(channel_path):
            raise ValueError(
                "PWM channel does not exist, check that the required modules are loaded."
            )

        pin_path = os.path.join(
            channel_path, self._pin_path.format(self._channel, self._pwmpin)
        )
        if not os.path.isdir(pin_path):
            try:
                with open(
                    os.path.join(channel_path, self._export_path), "w"
                ) as f_export:
                    f_export.write("%d\n" % self._pwmpin)
            except IOError as e:
                raise PWMError(e.errno, "Exporting PWM pin: " + e.strerror) from IOError

        # Look up the period, for fast duty cycle updates
        self._period = self._get_period()

        # set frequency
        self.frequency = freq
        # set duty
        self.duty_cycle = duty

        self._set_enabled(True)

    def deinit(self):
        """Deinit the sysfs PWM."""
        # pylint: disable=broad-except
        try:
            channel_path = os.path.join(
                self._sysfs_path, self._channel_path.format(self._channel)
            )

            if self._channel is not None:
                # self.duty_cycle = 0
                self._set_enabled(False)  # make to disable before unexport
                try:
                    # unexport_path = os.path.join(channel_path, self._unexport_path)
                    with open(
                        os.path.join(channel_path, self._unexport_path), "w"
                    ) as f_unexport:
                        f_unexport.write("%d\n" % self._pwmpin)
                except IOError as e:
                    raise PWMError(
                        e.errno, "Unexporting PWM pin: " + e.strerror
                    ) from IOError
        except Exception as e:
            # due to a race condition for which I have not yet been
            # able to find the root cause, deinit() often fails
            # but it does not effect future usage of the pwm pin
            print(
                "warning: failed to deinitialize pwm pin {0}:{1} due to: {2}\n".format(
                    self._channel, self._pwmpin, type(e).__name__
                )
            )
        finally:
            self._channel = None
            self._pwmpin = None
        # pylint: enable=broad-except

    def _is_deinited(self):
        if self._pwmpin is None:
            raise ValueError(
                "Object has been deinitialize and can no longer "
                "be used. Create a new object."
            )

    def _write_pin_attr(self, attr, value):
        # Make sure the pin is active
        self._is_deinited()

        path = os.path.join(
            self._sysfs_path,
            self._channel_path.format(self._channel),
            self._pin_path.format(self._channel, self._pwmpin),
            attr,
        )

        with open(path, "w") as f_attr:
            f_attr.write(value + "\n")

    def _read_pin_attr(self, attr):
        # Make sure the pin is active
        self._is_deinited()

        path = os.path.join(
            self._sysfs_path,
            self._channel_path.format(self._channel),
            self._pin_path.format(self._channel, self._pwmpin),
            attr,
        )

        with open(path, "r") as f_attr:
            return f_attr.read().strip()

    # Mutable properties

    def _get_period(self):
        period_ns = self._read_pin_attr(self._pin_period_path)
        try:
            period_ns = int(period_ns)
        except ValueError:
            raise PWMError(
                None, 'Unknown period value: "%s"' % period_ns
            ) from ValueError

        # Convert period from nanoseconds to seconds
        period = period_ns / 1e9

        # Update our cached period
        self._period = period

        return period

    def _set_period(self, period):
        if not isinstance(period, (int, float)):
            raise TypeError("Invalid period type, should be int or float.")

        # Convert period from seconds to integer nanoseconds
        period_ns = int(period * 1e9)

        self._write_pin_attr(self._pin_period_path, "{}".format(period_ns))

        # Update our cached period
        self._period = float(period)

    period = property(_get_period, _set_period)

    """Get or set the PWM's output period in seconds.

    Raises:
        PWMError: if an I/O or OS error occurs.
        TypeError: if value type is not int or float.

    :type: int, float
    """

    def _get_duty_cycle(self):
        duty_cycle_ns = self._read_pin_attr(self._pin_duty_cycle_path)
        try:
            duty_cycle_ns = int(duty_cycle_ns)
        except ValueError:
            raise PWMError(
                None, 'Unknown duty cycle value: "%s"' % duty_cycle_ns
            ) from ValueError

        # Convert duty cycle from nanoseconds to seconds
        duty_cycle = duty_cycle_ns / 1e9

        # Convert duty cycle to ratio from 0.0 to 1.0
        duty_cycle = duty_cycle / self._period

        # convert to 16-bit
        duty_cycle = int(duty_cycle * 65535)
        return duty_cycle

    def _set_duty_cycle(self, duty_cycle):
        if not isinstance(duty_cycle, (int, float)):
            raise TypeError("Invalid duty cycle type, should be int or float.")

        # convert from 16-bit
        duty_cycle /= 65535.0
        if not 0.0 <= duty_cycle <= 1.0:
            raise ValueError("Invalid duty cycle value, should be between 0.0 and 1.0.")

        # Convert duty cycle from ratio to seconds
        duty_cycle = duty_cycle * self._period

        # Convert duty cycle from seconds to integer nanoseconds
        duty_cycle_ns = int(duty_cycle * 1e9)

        self._write_pin_attr(self._pin_duty_cycle_path, "{}".format(duty_cycle_ns))

    duty_cycle = property(_get_duty_cycle, _set_duty_cycle)
    """Get or set the PWM's output duty cycle as a ratio from 0.0 to 1.0.

    Raises:
        PWMError: if an I/O or OS error occurs.
        TypeError: if value type is not int or float.
        ValueError: if value is out of bounds of 0.0 to 1.0.

    :type: int, float
    """

    def _get_frequency(self):
        return 1.0 / self._get_period()

    def _set_frequency(self, frequency):
        if not isinstance(frequency, (int, float)):
            raise TypeError("Invalid frequency type, should be int or float.")

        self._set_period(1.0 / frequency)

    frequency = property(_get_frequency, _set_frequency)
    """Get or set the PWM's output frequency in Hertz.

    Raises:
        PWMError: if an I/O or OS error occurs.
        TypeError: if value type is not int or float.

    :type: int, float
    """

    def _get_enabled(self):
        enabled = self._read_pin_attr(self._pin_enable_path)

        if enabled == "1":
            return True
        if enabled == "0":
            return False

        raise PWMError(None, 'Unknown enabled value: "%s"' % enabled)

    def _set_enabled(self, value):
        """Get or set the PWM's output enabled state.

        Raises:
            PWMError: if an I/O or OS error occurs.
            TypeError: if value type is not bool.

        :type: bool
        """
        if not isinstance(value, bool):
            raise TypeError("Invalid enabled type, should be string.")

        self._write_pin_attr(self._pin_enable_path, "1" if value else "0")

    # String representation

    def __str__(self):
        return "PWM%d, pin %s (freq=%f Hz, duty_cycle=%f%%)" % (
            self._channel,
            self._pin,
            self.frequency,
            self.duty_cycle * 100,
        )



"""Chip Definition for MCP2221"""

import os
import time
import hid

# Here if you need it
MCP2221_HID_DELAY = float(os.environ.get("BLINKA_MCP2221_HID_DELAY", 0))
# Use to set delay between reset and device reopen. if negative, don't reset at all
MCP2221_RESET_DELAY = float(os.environ.get("BLINKA_MCP2221_RESET_DELAY", 0.5))

# from the C driver
# http://ww1.microchip.com/downloads/en/DeviceDoc/mcp2221_0_1.tar.gz
# others (???) determined during driver developement
RESP_ERR_NOERR = 0x00
RESP_ADDR_NACK = 0x25
RESP_READ_ERR = 0x7F
RESP_READ_COMPL = 0x55
RESP_READ_PARTIAL = 0x54  # ???
RESP_I2C_IDLE = 0x00
RESP_I2C_START_TOUT = 0x12
RESP_I2C_RSTART_TOUT = 0x17
RESP_I2C_WRADDRL_TOUT = 0x23
RESP_I2C_WRADDRL_WSEND = 0x21
RESP_I2C_WRADDRL_NACK = 0x25
RESP_I2C_WRDATA_TOUT = 0x44
RESP_I2C_RDDATA_TOUT = 0x52
RESP_I2C_STOP_TOUT = 0x62

RESP_I2C_MOREDATA = 0x43  # ???
RESP_I2C_PARTIALDATA = 0x41  # ???
RESP_I2C_WRITINGNOSTOP = 0x45  # ???

MCP2221_RETRY_MAX = 50
MCP2221_MAX_I2C_DATA_LEN = 60
MASK_ADDR_NACK = 0x40


class MCP2221:
    """MCP2221 Device Class Definition"""

    VID = 0x04D8
    PID = 0x00DD

    GP_GPIO = 0b000
    GP_DEDICATED = 0b001
    GP_ALT0 = 0b010
    GP_ALT1 = 0b011
    GP_ALT2 = 0b100

    def __init__(self):
        self._hid = hid.device()
        self._hid.open(MCP2221.VID, MCP2221.PID)
        if MCP2221_RESET_DELAY >= 0:
            self._reset()
        self._gp_config = [0x07] * 4  # "don't care" initial value
        for pin in range(4):
            self.gp_set_mode(pin, self.GP_GPIO)  # set to GPIO mode
            self.gpio_set_direction(pin, 1)  # set to INPUT

    def _hid_xfer(self, report, response=True):
        """Perform HID Transfer"""
        # first byte is report ID, which =0 for MCP2221
        # remaing bytes = 64 byte report data
        # https://github.com/libusb/hidapi/blob/083223e77952e1ef57e6b77796536a3359c1b2a3/hidapi/hidapi.h#L185
        self._hid.write(b"\0" + report + b"\0" * (64 - len(report)))
        time.sleep(MCP2221_HID_DELAY)
        if response:
            # return is 64 byte response report
            return self._hid.read(64)
        return None

    # ----------------------------------------------------------------
    # MISC
    # ----------------------------------------------------------------
    def gp_get_mode(self, pin):
        """Get Current Pin Mode"""
        return self._hid_xfer(b"\x61")[22 + pin] & 0x07

    def gp_set_mode(self, pin, mode):
        """Set Current Pin Mode"""
        # already set to that mode?
        mode &= 0x07
        if mode == (self._gp_config[pin] & 0x07):
            return
        # update GP mode for pin
        self._gp_config[pin] = mode
        # empty report, this is safe since 0's = no change
        report = bytearray(b"\x60" + b"\x00" * 63)
        # set the alter GP flag byte
        report[7] = 0xFF
        # add GP setttings
        report[8] = self._gp_config[0]
        report[9] = self._gp_config[1]
        report[10] = self._gp_config[2]
        report[11] = self._gp_config[3]
        # and make it so
        self._hid_xfer(report)

    def _pretty_report(self, register):
        report = self._hid_xfer(register)
        print("     0  1  2  3  4  5  6  7  8  9")
        index = 0
        for row in range(7):
            print("{} : ".format(row), end="")
            for _ in range(10):
                print("{:02x} ".format(report[index]), end="")
                index += 1
                if index > 63:
                    break
            print()

    def _status_dump(self):
        self._pretty_report(b"\x10")

    def _sram_dump(self):
        self._pretty_report(b"\x61")

    def _reset(self):
        self._hid_xfer(b"\x70\xAB\xCD\xEF", response=False)
        time.sleep(MCP2221_RESET_DELAY)
        start = time.monotonic()
        while time.monotonic() - start < 5:
            try:
                self._hid.open(MCP2221.VID, MCP2221.PID)
            except OSError:
                # try again
                time.sleep(0.1)
                continue
            return
        raise OSError("open failed")

    # ----------------------------------------------------------------
    # GPIO
    # ----------------------------------------------------------------
    def gpio_set_direction(self, pin, mode):
        """Set Current GPIO Pin Direction"""
        if mode:
            # set bit 3 for INPUT
            self._gp_config[pin] |= 1 << 3
        else:
            # clear bit 3 for OUTPUT
            self._gp_config[pin] &= ~(1 << 3)
        report = bytearray(b"\x50" + b"\x00" * 63)  # empty set GPIO report
        offset = 4 * (pin + 1)
        report[offset] = 0x01  # set pin direction
        report[offset + 1] = mode  # to this
        self._hid_xfer(report)

    def gpio_set_pin(self, pin, value):
        """Set Current GPIO Pin Value"""
        if value:
            # set bit 4
            self._gp_config[pin] |= 1 << 4
        else:
            # clear bit 4
            self._gp_config[pin] &= ~(1 << 4)
        report = bytearray(b"\x50" + b"\x00" * 63)  # empty set GPIO report
        offset = 2 + 4 * pin
        report[offset] = 0x01  # set pin value
        report[offset + 1] = value  # to this
        self._hid_xfer(report)

    def gpio_get_pin(self, pin):
        """Get Current GPIO Pin Value"""
        resp = self._hid_xfer(b"\x51")
        offset = 2 + 2 * pin
        if resp[offset] == 0xEE:
            raise RuntimeError("Pin is not set for GPIO operation.")
        return resp[offset]

    # ----------------------------------------------------------------
    # I2C
    # ----------------------------------------------------------------
    def _i2c_status(self):
        resp = self._hid_xfer(b"\x10")
        if resp[1] != 0:
            raise RuntimeError("Couldn't get I2C status")
        return resp

    def _i2c_state(self):
        return self._i2c_status()[8]

    def _i2c_cancel(self):
        resp = self._hid_xfer(b"\x10\x00\x10")
        if resp[1] != 0x00:
            raise RuntimeError("Couldn't cancel I2C")
        if resp[2] == 0x10:
            # bus release will need "a few hundred microseconds"
            time.sleep(0.001)

    # pylint: disable=too-many-arguments
    def _i2c_write(self, cmd, address, buffer, start=0, end=None):
        if self._i2c_state() != 0x00:
            self._i2c_cancel()

        end = end if end else len(buffer)
        length = end - start
        retries = 0

        while (end - start) > 0:
            chunk = min(end - start, MCP2221_MAX_I2C_DATA_LEN)
            # write out current chunk
            resp = self._hid_xfer(
                bytes([cmd, length & 0xFF, (length >> 8) & 0xFF, address << 1])
                + buffer[start : (start + chunk)]
            )
            # check for success
            if resp[1] != 0x00:
                if resp[2] in (
                    RESP_I2C_START_TOUT,
                    RESP_I2C_WRADDRL_TOUT,
                    RESP_I2C_WRADDRL_NACK,
                    RESP_I2C_WRDATA_TOUT,
                    RESP_I2C_STOP_TOUT,
                ):
                    raise RuntimeError("Unrecoverable I2C state failure")
                retries += 1
                if retries >= MCP2221_RETRY_MAX:
                    raise RuntimeError("I2C write error, max retries reached.")
                time.sleep(0.001)
                continue  # try again
            # yay chunk sent!
            while self._i2c_state() == RESP_I2C_PARTIALDATA:
                time.sleep(0.001)
            start += chunk
            retries = 0

        # check status in another loop
        for _ in range(MCP2221_RETRY_MAX):
            status = self._i2c_status()
            if status[20] & MASK_ADDR_NACK:
                raise RuntimeError("I2C slave address was NACK'd")
            usb_cmd_status = status[8]
            if usb_cmd_status == 0:
                break
            if usb_cmd_status == RESP_I2C_WRITINGNOSTOP and cmd == 0x94:
                break  # this is OK too!
            if usb_cmd_status in (
                RESP_I2C_START_TOUT,
                RESP_I2C_WRADDRL_TOUT,
                RESP_I2C_WRADDRL_NACK,
                RESP_I2C_WRDATA_TOUT,
                RESP_I2C_STOP_TOUT,
            ):
                raise RuntimeError("Unrecoverable I2C state failure")
            time.sleep(0.001)
        else:
            raise RuntimeError("I2C write error: max retries reached.")
        # whew success!

    def _i2c_read(self, cmd, address, buffer, start=0, end=None):
        if self._i2c_state() not in (RESP_I2C_WRITINGNOSTOP, 0):
            self._i2c_cancel()

        end = end if end else len(buffer)
        length = end - start

        # tell it we want to read
        resp = self._hid_xfer(
            bytes([cmd, length & 0xFF, (length >> 8) & 0xFF, (address << 1) | 0x01])
        )

        # check for success
        if resp[1] != 0x00:
            raise RuntimeError("Unrecoverable I2C read failure")

        # and now the read part
        while (end - start) > 0:
            for _ in range(MCP2221_RETRY_MAX):
                # the actual read
                resp = self._hid_xfer(b"\x40")
                # check for success
                if resp[1] == RESP_I2C_PARTIALDATA:
                    time.sleep(0.001)
                    continue
                if resp[1] != 0x00:
                    raise RuntimeError("Unrecoverable I2C read failure")
                if resp[2] == RESP_ADDR_NACK:
                    raise RuntimeError("I2C NACK")
                if resp[3] == 0x00 and resp[2] == 0x00:
                    break
                if resp[3] == RESP_READ_ERR:
                    time.sleep(0.001)
                    continue
                if resp[2] in (RESP_READ_COMPL, RESP_READ_PARTIAL):
                    break
            else:
                raise RuntimeError("I2C read error: max retries reached.")

            # move data into buffer
            chunk = min(end - start, 60)
            for i, k in enumerate(range(start, start + chunk)):
                buffer[k] = resp[4 + i]
            start += chunk

    # pylint: enable=too-many-arguments

    def _i2c_configure(self, baudrate=100000):
        """Configure I2C"""
        self._hid_xfer(
            bytes(
                [
                    0x10,  # set parameters
                    0x00,  # don't care
                    0x00,  # no effect
                    0x20,  # next byte is clock divider
                    12000000 // baudrate - 3,
                ]
            )
        )

    def i2c_writeto(self, address, buffer, *, start=0, end=None):
        """Write data from the buffer to an address"""
        self._i2c_write(0x90, address, buffer, start, end)

    def i2c_readfrom_into(self, address, buffer, *, start=0, end=None):
        """Read data from an address and into the buffer"""
        self._i2c_read(0x91, address, buffer, start, end)

    def i2c_writeto_then_readfrom(
        self,
        address,
        out_buffer,
        in_buffer,
        *,
        out_start=0,
        out_end=None,
        in_start=0,
        in_end=None
    ):
        """Write data from buffer_out to an address and then
        read data from an address and into buffer_in
        """
        self._i2c_write(0x94, address, out_buffer, out_start, out_end)
        self._i2c_read(0x93, address, in_buffer, in_start, in_end)

    def i2c_scan(self, *, start=0, end=0x79):
        """Perform an I2C Device Scan"""
        found = []
        for addr in range(start, end + 1):
            # try a write
            try:
                self.i2c_writeto(addr, b"\x00")
            except RuntimeError:  # no reply!
                continue
            # store if success
            found.append(addr)
        return found

    # ----------------------------------------------------------------
    # ADC
    # ----------------------------------------------------------------
    def adc_configure(self, vref=0):
        """Configure the Analog-to-Digital Converter"""
        report = bytearray(b"\x60" + b"\x00" * 63)
        report[5] = 1 << 7 | (vref & 0b111)
        self._hid_xfer(report)

    def adc_read(self, pin):
        """Read from the Analog-to-Digital Converter"""
        resp = self._hid_xfer(b"\x10")
        return resp[49 + 2 * pin] << 8 | resp[48 + 2 * pin]

    # ----------------------------------------------------------------
    # DAC
    # ----------------------------------------------------------------
    def dac_configure(self, vref=0):
        """Configure the Digital-to-Analog Converter"""
        report = bytearray(b"\x60" + b"\x00" * 63)
        report[3] = 1 << 7 | (vref & 0b111)
        self._hid_xfer(report)

    # pylint: disable=unused-argument
    def dac_write(self, pin, value):
        """Write to the Digital-to-Analog Converter"""
        report = bytearray(b"\x60" + b"\x00" * 63)
        report[4] = 1 << 7 | (value & 0b11111)
        self._hid_xfer(report)

    # pylint: enable=unused-argument


mcp2221 = MCP2221()



"""
`pulseio` - Pulse Width Modulation input and output control
===========================================================
See `CircuitPython:pulseio` in CircuitPython for more details.
Not supported by all boards.

* Author(s): Melissa LeBlanc-Williams
"""

import sys

from adafruit_blinka.agnostic import detector

# pylint: disable=unused-import

if detector.board.any_raspberry_pi:
    from adafruit_blinka.microcontroller.bcm283x.pulseio.PulseIn import PulseIn
    from adafruit_blinka.microcontroller.bcm283x.pulseio.PWMOut import PWMOut
elif detector.board.any_coral_board:
    from adafruit_blinka.microcontroller.generic_linux.sysfs_pwmout import PWMOut
elif detector.board.any_giant_board:
    from adafruit_blinka.microcontroller.generic_linux.sysfs_pwmout import PWMOut
elif detector.board.any_beaglebone:
    from adafruit_blinka.microcontroller.am335x.sysfs_pwmout import PWMOut
elif detector.board.any_rock_pi_board:
    from adafruit_blinka.microcontroller.generic_linux.sysfs_pwmout import PWMOut
elif detector.board.binho_nova:
    from adafruit_blinka.microcontroller.nova.pwmout import PWMOut
elif detector.board.greatfet_one:
    from adafruit_blinka.microcontroller.nxp_lpc4330.pwmout import PWMOut
elif detector.board.any_lubancat:
    from adafruit_blinka.microcontroller.generic_linux.sysfs_pwmout import PWMOut
elif "sphinx" in sys.modules:
    pass
else:
    raise NotImplementedError("pulseio not supported for this board.")



"""Pin definitions for the Dragonboard 410c."""

from adafruit_blinka.microcontroller.snapdragon.apq8016 import pin

GPIO_A = pin.GPIO_36
GPIO_B = pin.GPIO_12
GPIO_C = pin.GPIO_13
GPIO_D = pin.GPIO_69
GPIO_E = pin.GPIO_115
GPIO_F = pin.PM_MPP_4
GPIO_G = pin.GPIO_24
GPIO_H = pin.GPIO_25
GPIO_I = pin.GPIO_35
GPIO_J = pin.GPIO_34
GPIO_K = pin.GPIO_28
GPIO_L = pin.GPIO_33

GPIO_36 = pin.GPIO_36
GPIO_12 = pin.GPIO_12
GPIO_13 = pin.GPIO_13
GPIO_69 = pin.GPIO_69
GPIO_115 = pin.GPIO_115
GPIO_4 = pin.PM_MPP_4
GPIO_24 = pin.GPIO_24
GPIO_25 = pin.GPIO_25
GPIO_35 = pin.GPIO_35
GPIO_34 = pin.GPIO_34
GPIO_28 = pin.GPIO_28
GPIO_33 = pin.GPIO_33

SDA = pin.I2C0_SDA
SCL = pin.I2C0_SCL

I2C0_SDA = pin.I2C0_SDA
I2C0_SCL = pin.I2C0_SCL
I2C1_SDA = pin.I2C1_SDA
I2C1_SCL = pin.I2C1_SCL

SCLK = pin.SPI0_SCLK
MOSI = pin.SPI0_MOSI
MISO = pin.SPI0_MISO
SPI_CS = pin.SPI0_CS



"""Tegra T210 pin names"""

import atexit
import Jetson.GPIO as GPIO

GPIO.setmode(GPIO.TEGRA_SOC)
GPIO.setwarnings(False)  # shh!


class Pin:
    """Pins dont exist in CPython so...lets make our own!"""

    IN = 0
    OUT = 1
    LOW = 0
    HIGH = 1
    PULL_NONE = 0
    PULL_UP = 1
    PULL_DOWN = 2

    id = None
    _value = LOW
    _mode = IN

    def __init__(self, bcm_number):
        self.id = bcm_number

    def __repr__(self):
        return str(self.id)

    def __eq__(self, other):
        return self.id == other

    def init(self, mode=IN, pull=None):
        """Initialize the Pin"""
        if mode is not None:
            if mode == self.IN:
                self._mode = self.IN
                GPIO.setup(self.id, GPIO.IN)
            elif mode == self.OUT:
                self._mode = self.OUT
                GPIO.setup(self.id, GPIO.OUT)
            else:
                raise RuntimeError("Invalid mode for pin: %s" % self.id)
        if pull is not None:
            if self._mode != self.IN:
                raise RuntimeError("Cannot set pull resistor on output")
            if pull == self.PULL_UP:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            elif pull == self.PULL_DOWN:
                GPIO.setup(self.id, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            else:
                raise RuntimeError("Invalid pull for pin: %s" % self.id)

    def value(self, val=None):
        """Set or return the Pin Value"""
        if val is not None:
            if val == self.LOW:
                self._value = val
                GPIO.output(self.id, val)
                return None
            if val == self.HIGH:
                self._value = val
                GPIO.output(self.id, val)
                return None
            raise RuntimeError("Invalid value for pin")
        return GPIO.input(self.id)

    # pylint: disable=no-method-argument
    @atexit.register
    def cleanup():
        """Clean up pins"""
        print("Exiting... \nCleaning up pins")
        GPIO.cleanup()

    # pylint: enable=no-method-argument


# Cannot be used as GPIO
SDA = Pin("GEN1_I2C_SDA")
SCL = Pin("GEN1_I2C_SCL")
SDA_1 = Pin("GEN2_I2C_SDA")
SCL_1 = Pin("GEN2_I2C_SCL")

# These pins are native to TX1
BB03 = Pin("GPIO_X1_AUD")
X02 = Pin("MOTION_INT")
H07 = Pin("AP_WAKE_NFC")
E04 = Pin("DMIC3_CLK")
U03 = Pin("UART1_CTS")
U02 = Pin("UART1_RTS")
B03 = Pin("DAP1_SCLK")
B00 = Pin("DAP1_FS")
B01 = Pin("DAP1_DIN")
B02 = Pin("DAP1_DOUT")
P17 = Pin("GPIO_EXP_P17")
E05 = Pin("DMIC3_DAT")
X00 = Pin("MODEM_WAKE_AP")
P16 = Pin("GPIO_EXP_P16")
X03 = Pin("ALS_PROX_INT")

# These pins are native to NANO
S05 = Pin("CAM_AF_EN")
Z00 = Pin("GPIO_PZ0")
V00 = Pin("LCD_BL_PW")
G03 = Pin("UART2_CTS")
G02 = Pin("UART2_RTS")
J07 = Pin("DAP4_SCLK")
J04 = Pin("DAP4_FS")
J05 = Pin("DAP4_DIN")
J06 = Pin("DAP4_DOUT")
Y02 = Pin("LCD_TE")
DD00 = Pin("SPI2_CS1")
B07 = Pin("SPI2_CS0")
B05 = Pin("SPI2_MISO")
B04 = Pin("SPI2_MOSI")
B06 = Pin("SPI2_SCK")

# These pins are shared across T210
BB00 = Pin("AUD_MCLK")
C04 = Pin("SPI1_CS1")
C03 = Pin("SPI1_CS0")
C01 = Pin("SPI1_MISO")
C00 = Pin("SPI1_MOSI")
C02 = Pin("SPI1_SCK")
E06 = Pin("GPIO_PE6")

i2cPorts = (
    (0, SCL, SDA),
    (1, SCL_1, SDA_1),
)

# ordered as spiId, sckId, mosiId, misoId
spiPorts = ((0, C02, C00, C01), (1, B06, B04, B05))



"""Analog in demo"""
import time
import board
from analogio import AnalogIn

analog_in = AnalogIn(board.A1)


def get_voltage(pin):
    return (pin.value * 3.3) / 4096


while True:
    print((get_voltage(analog_in),))
    time.sleep(0.1)



"""
Much code from https://github.com/vsergeev/python-periphery/blob/master/periphery/gpio.py
Copyright (c) 2015-2019 vsergeev / Ivan (Vanya) A. Sergeev
License: MIT
"""

import os
import errno
import time

# pylint: disable=unnecessary-pass
class GPIOError(IOError):
    """Base class for GPIO errors."""

    pass


# pylint: enable=unnecessary-pass


class Pin:
    """SysFS GPIO Pin Class"""

    # Number of retries to check for GPIO export or direction write on open
    GPIO_OPEN_RETRIES = 10
    # Delay between check for GPIO export or direction write on open (100ms)
    GPIO_OPEN_DELAY = 0.1

    IN = "in"
    OUT = "out"
    LOW = 0
    HIGH = 1
    PULL_NONE = 0
    PULL_UP = 1
    PULL_DOWN = 2

    id = None
    _value = LOW
    _mode = IN

    # Sysfs paths
    _sysfs_path = "/sys/class/gpio/"
    _channel_path = "gpiochip{}"

    # Channel paths
    _export_path = "export"
    _unexport_path = "unexport"
    _pin_path = "gpio{}"

    def __init__(self, pin_id):
        """Instantiate a Pin object and open the sysfs GPIO with the specified
        pin number.

        Args:
            pin_id (int): GPIO pin number.

        Returns:
            SysfsGPIO: GPIO object.

        Raises:
            GPIOError: if an I/O or OS error occurs.
            TypeError: if `line` or `direction`  types are invalid.
            ValueError: if `direction` value is invalid.
            TimeoutError: if waiting for GPIO export times out.

        """

        if not isinstance(pin_id, int):
            raise TypeError("Invalid Pin ID, should be integer.")

        self.id = pin_id
        self._fd = None
        self._line = None
        self._path = None

    def __del__(self):
        self._close()

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self._close()

    def init(self, mode=IN, pull=None):
        """Initialize the Pin"""
        if mode is not None:
            if mode == self.IN:
                self._mode = self.IN
                self._open(self.IN)
            elif mode == self.OUT:
                self._mode = self.OUT
                self._open(self.OUT)
            else:
                raise RuntimeError("Invalid mode for pin: %s" % self.id)

            if pull is not None:
                if pull == self.PULL_UP:
                    raise NotImplementedError(
                        "Internal pullups not supported in periphery, "
                        "use physical resistor instead!"
                    )
                if pull == self.PULL_DOWN:
                    raise NotImplementedError(
                        "Internal pulldowns not supported in periphery, "
                        "use physical resistor instead!"
                    )
                raise RuntimeError("Invalid pull for pin: %s" % self.id)

    def value(self, val=None):
        """Set or return the Pin Value"""
        if val is not None:
            if val == self.LOW:
                self._value = val
                self._write(False)
                return None
            if val == self.HIGH:
                self._value = val
                self._write(True)
                return None
            raise RuntimeError("Invalid value for pin")
        return self.HIGH if self._read() else self.LOW

    # pylint: disable=too-many-branches
    def _open(self, direction):
        if not isinstance(direction, str):
            raise TypeError("Invalid direction type, should be string.")
        if direction.lower() not in ["in", "out", "high", "low"]:
            raise ValueError('Invalid direction, can be: "in", "out", "high", "low".')
        gpio_path = "/sys/class/gpio/gpio{:d}".format(self.id)

        if not os.path.isdir(gpio_path):
            # Export the line
            try:
                with open("/sys/class/gpio/export", "w") as f_export:
                    f_export.write("{:d}\n".format(self.id))
            except IOError as e:
                raise GPIOError(e.errno, "Exporting GPIO: " + e.strerror) from IOError

            # Loop until GPIO is exported
            exported = False
            for i in range(self.GPIO_OPEN_RETRIES):
                if os.path.isdir(gpio_path):
                    exported = True
                    break

                time.sleep(self.GPIO_OPEN_DELAY)

            if not exported:
                raise TimeoutError(
                    'Exporting GPIO: waiting for "{:s}" timed out'.format(gpio_path)
                )

            # Write direction, looping in case of EACCES errors due to delayed udev
            # permission rule application after export
            for i in range(self.GPIO_OPEN_RETRIES):
                try:
                    with open(os.path.join(gpio_path, "direction"), "w") as f_direction:
                        f_direction.write(direction.lower() + "\n")
                    break
                except IOError as e:
                    if e.errno != errno.EACCES or (
                        e.errno == errno.EACCES and i == self.GPIO_OPEN_RETRIES - 1
                    ):
                        raise GPIOError(
                            e.errno, "Setting GPIO direction: " + e.strerror
                        ) from IOError

                time.sleep(self.GPIO_OPEN_DELAY)
        else:
            # Write direction
            try:
                with open(os.path.join(gpio_path, "direction"), "w") as f_direction:
                    f_direction.write(direction.lower() + "\n")
            except IOError as e:
                raise GPIOError(
                    e.errno, "Setting GPIO direction: " + e.strerror
                ) from IOError

        # Open value
        try:
            self._fd = os.open(os.path.join(gpio_path, "value"), os.O_RDWR)
        except OSError as e:
            raise GPIOError(e.errno, "Opening GPIO: " + e.strerror) from OSError

        self._path = gpio_path

    # pylint: enable=too-many-branches

    def _close(self):
        if self._fd is None:
            return

        try:
            os.close(self._fd)
        except OSError as e:
            raise GPIOError(e.errno, "Closing GPIO: " + e.strerror) from OSError

        self._fd = None

        # Unexport the line
        try:
            unexport_fd = os.open("/sys/class/gpio/unexport", os.O_WRONLY)
            os.write(unexport_fd, "{:d}\n".format(self.id).encode())
            os.close(unexport_fd)
        except OSError as e:
            raise GPIOError(e.errno, "Unexporting GPIO: " + e.strerror) from OSError

    def _read(self):
        # Read value
        try:
            buf = os.read(self._fd, 2)
        except OSError as e:
            raise GPIOError(e.errno, "Reading GPIO: " + e.strerror) from OSError

        # Rewind
        try:
            os.lseek(self._fd, 0, os.SEEK_SET)
        except OSError as e:
            raise GPIOError(e.errno, "Rewinding GPIO: " + e.strerror) from OSError

        if buf[0] == b"0"[0]:
            return False
        if buf[0] == b"1"[0]:
            return True

        raise GPIOError(None, "Unknown GPIO value: {}".format(buf))

    def _write(self, value):
        if not isinstance(value, bool):
            raise TypeError("Invalid value type, should be bool.")

        # Write value
        try:
            if value:
                os.write(self._fd, b"1\n")
            else:
                os.write(self._fd, b"0\n")
        except OSError as e:
            raise GPIOError(e.errno, "Writing GPIO: " + e.strerror) from OSError

        # Rewind
        try:
            os.lseek(self._fd, 0, os.SEEK_SET)
        except OSError as e:
            raise GPIOError(e.errno, "Rewinding GPIO: " + e.strerror) from OSError

    @property
    def chip_name(self):
        """Return the Chip Name"""
        gpio_path = os.path.join(self._path, "device")

        gpiochip_path = os.readlink(gpio_path)

        if "/" not in gpiochip_path:
            raise GPIOError(
                None,
                'Reading gpiochip name: invalid device symlink "{:s}"'.format(
                    gpiochip_path
                ),
            )

        return gpiochip_path.split("/")[-1]

    @property
    def chip_label(self):
        """Return the Chip Label"""
        gpio_path = "/sys/class/gpio/{:s}/label".format(self.chip_name)

        try:
            with open(gpio_path, "r") as f_label:
                label = f_label.read()
        except (GPIOError, IOError) as e:
            if isinstance(e, IOError):
                raise GPIOError(
                    e.errno, "Reading gpiochip label: " + e.strerror
                ) from IOError

            raise GPIOError(
                None, "Reading gpiochip label: " + e.strerror
            ) from GPIOError

        return label.strip()

    # Mutable properties

    def _get_direction(self):
        # Read direction
        try:
            with open(os.path.join(self._path, "direction"), "r") as f_direction:
                direction = f_direction.read()
        except IOError as e:
            raise GPIOError(
                e.errno, "Getting GPIO direction: " + e.strerror
            ) from IOError

        return direction.strip()

    def _set_direction(self, direction):
        if not isinstance(direction, str):
            raise TypeError("Invalid direction type, should be string.")
        if direction.lower() not in ["in", "out", "high", "low"]:
            raise ValueError('Invalid direction, can be: "in", "out", "high", "low".')

        # Write direction
        try:
            with open(os.path.join(self._path, "direction"), "w") as f_direction:
                f_direction.write(direction.lower() + "\n")
        except IOError as e:
            raise GPIOError(
                e.errno, "Setting GPIO direction: " + e.strerror
            ) from IOError

    direction = property(_get_direction, _set_direction)

    def __str__(self):
        try:
            str_direction = self.direction
        except GPIOError:
            str_direction = "<error>"

        try:
            str_chip_name = self.chip_name
        except GPIOError:
            str_chip_name = "<error>"

        try:
            str_chip_label = self.chip_label
        except GPIOError:
            str_chip_label = "<error>"

        output = "Pin {:d} (dev={:s}, fd={:d}, dir={:s}, chip_name='{:s}', chip_label='{:s}')"
        return output.format(
            self.id, self._path, self._fd, str_direction, str_chip_name, str_chip_label
        )



"""Pin definitions for the Feather Huzzah."""

from adafruit_blinka.microcontroller.esp8266 import pin

# TODO need equiv of INPUT_PULL_DOWN_16 ?
# See https://tttapa.github.io/ESP8266/Chap04%20-%20Microcontroller.html

GPIO0 = pin.GPIO0
GPIO1 = pin.GPIO1
GPIO2 = pin.GPIO2
GPIO3 = pin.GPIO3
GPIO4 = pin.GPIO4
GPIO5 = pin.GPIO5
GPIO12 = pin.GPIO12
GPIO13 = pin.GPIO13
GPIO14 = pin.GPIO14
GPIO15 = pin.GPIO15
GPIO16 = pin.GPIO16

ADC = pin.TOUT

MISO = GPIO12
MOSI = GPIO13
SCK = GPIO14

RX = GPIO3
TX = GPIO1

SDA = GPIO4
SCL = GPIO5
